import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scanpy.get import obs_df
from matplotlib.cm import get_cmap
import itertools
import logging
from matplotlib import use
import muon as mu
import matplotlib
from scipy.sparse import issparse
from scvi import REGISTRY_KEYS

use('Agg')
plt.ioff()


def cell2loc_plot_history(model,fig_path, iter_start=0, iter_end=-1, ax=None):
# Adapted from: https://github.com/BayraktarLab/cell2location/blob/master/cell2location/models/base/_pyro_mixin.py#L407
    if ax is None:
        ax = plt.gca()
    if iter_end == -1:
        iter_end = len(model.history_["elbo_train"])

    ax.plot(
        np.array(model.history_["elbo_train"].index[iter_start:iter_end]),
        np.array(model.history_["elbo_train"].values.flatten())[iter_start:iter_end],
        label="train",
    )
    ax.legend()
    ax.set_xlim(0, len(model.history_["elbo_train"]))
    ax.set_xlabel("Training epochs")
    ax.set_ylabel("-ELBO loss")
    plt.tight_layout()
    plt.savefig(fig_path)


def cell2loc_plot_QC_reconstr(model, fig_path, summary_name: str = "means", use_n_obs: int = 1000):
# Adapted from: https://github.com/BayraktarLab/cell2location/blob/master/cell2location/models/base/_pyro_mixin.py#L544

    if getattr(model, "samples", False) is False:
        raise RuntimeError("self.samples is missing, please run self.export_posterior() first")
    if use_n_obs is not None:
        ind_x = np.random.choice(
            model.adata_manager.adata.n_obs, np.min((use_n_obs, model.adata.n_obs)), replace=False
        )
    else:
        ind_x = None

    model.expected_nb_param = model.module.model.compute_expected(
        model.samples[f"post_sample_{summary_name}"], model.adata_manager, ind_x=ind_x
    )
    x_data = model.adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY)[ind_x, :]
    if issparse(x_data):
        x_data = np.asarray(x_data.toarray())
    mu = model.expected_nb_param["mu"]
    plt.hist2d(
        np.log10(x_data.flatten() + 1),
        np.log10(mu.flatten() + 1),
        bins=50,
        norm=matplotlib.colors.LogNorm(),
    )
    plt.gca().set_aspect("equal", adjustable="box")
    plt.xlabel("Data, log10")
    plt.ylabel("Posterior expected value, log10")
    plt.title("Reconstruction accuracy")
    plt.tight_layout()
    plt.savefig(fig_path)


def cell2loc_plot_QC_reference(reference_model,fig_path_reconstr, fig_path_expr,
                               summary_name: str = "means",
                               use_n_obs: int = 1000,
                               scale_average_detection: bool = True,):
# Adapted from: https://github.com/BayraktarLab/cell2location/blob/master/cell2location/models/reference/_reference_model.py#L265

    # 1. plot reconstruction accuracy
    cell2loc_plot_QC_reconstr(model=reference_model,fig_path=fig_path_reconstr, summary_name=summary_name, use_n_obs=use_n_obs)

    # 2. plot estimated reference expression signatures (accounting for batch effect) compared to average expression in each cluster
    inf_aver = reference_model.samples[f"post_sample_{summary_name}"]["per_cluster_mu_fg"].T
    if scale_average_detection and ("detection_y_c" in list(reference_model.samples[f"post_sample_{summary_name}"].keys())):
        inf_aver = inf_aver * reference_model.samples[f"post_sample_{summary_name}"]["detection_y_c"].mean()
    aver = reference_model._compute_cluster_averages(key=REGISTRY_KEYS.LABELS_KEY)
    aver = aver[reference_model.factor_names_]

    plt.hist2d(
        np.log10(aver.values.flatten() + 1),
        np.log10(inf_aver.flatten() + 1),
        bins=50,
        norm=matplotlib.colors.LogNorm(),
    )
    plt.xlabel("Mean expression for every gene in every cluster")
    plt.ylabel("Estimated expression for every gene in every cluster")
    plt.savefig(fig_path_expr)


def plot_lsi_corr(corr_df, tech_covariates=['n_genes_by_counts', 'total_counts'], filename="corr_plot.png"):
    plt.figure(figsize=(10, 6))  # Optional: You can adjust the figure size

    for tech in tech_covariates:
        subset = corr_df[corr_df['Tech_Covariate'] == tech]
        plt.scatter(subset['LSI_Component'], subset['Correlation'], label=f'{tech} Correlation')

    plt.xlabel('LSI Component')
    plt.ylabel('Correlation Value')
    plt.title('Correlation between LSI Components and Technical Covariates')
    plt.legend()
    plt.xticks(rotation=45)
    
    plt.savefig(filename)  # Saving the plot to the specified file
    plt.close()  # Optional: Close the figure after saving to free up memory


def scatter_one(group_choice, col_choice, plot_df, axs=None, colour="#1f77b4", title=""):
    plot_df['plot_col'] = plot_df[col_choice]
    # highlight one choice
    other_choices = plot_df['plot_col'].unique().tolist()
    other_choices.remove(group_choice)
    # " nan" is an essential  cheat to get the nans to plot at the back
    plot_df['plot_col'] = plot_df['plot_col'].replace(other_choices, " nan")
    plot_df['plot_col'] = plot_df['plot_col'].astype("category")
    plot_df['plot_col'] = plot_df['plot_col'].cat.set_categories([group_choice, " nan"], ordered=True)
    # reorder plot so nans are plotted first
    plot_df = plot_df.sort_values(by='plot_col', ascending=False)
    # do plot
    if axs is None:
        fig, axs = plt.subplots(figsize=(3, 3), facecolor='w', edgecolor='k')
    sns.scatterplot(data=plot_df, x=plot_df.columns[0], y=plot_df.columns[1], hue="plot_col", palette=[colour, "#dddddd"],
                    s=2, alpha=.3, linewidth=0,
                    legend=False, ax=axs)
    axs.axis("off")
    axs.set_title(title)
    return axs


def batch_scatter_two_var(plot_df, method, batch, palette_choice=None):
    """Plots facetted umaps, with each group within the method and batch highlighted as foreground.
    Args:
        plot_df (pd.DataFrame): pandas dataframe contianin umap coordinates plus method and batch columns.
        method (str): method column in plot_df (this will be plotted one category per column)
        batch (str): batch column in plot_df (this will be plotted one category per row)
        palette_choice (list): List of colors to plot each rows foreground. Defaults to None.
    Returns:
        fig, ax : matplotlib subplots figure
    """
    plot_df = plot_df[['umap_1', 'umap_2', method, batch]]
    plot_df = plot_df.dropna()
    # n_vars = len(df[var_choice].unique())
    method_choices = plot_df[method].cat.categories.tolist()
    logging.debug(method_choices)
    group_choices = plot_df[batch].unique()
    logging.debug(group_choices)
    nrows = len(group_choices)
    ncols = len(method_choices)
    if nrows > 40:
        logging.info("skipping facet plot as too many variables: %s" % batch)
        fig = None
        axs = None
    else:
        fig, axs = plt.subplots(nrows, ncols, figsize=(4*ncols, 3*nrows), facecolor='w', edgecolor='k')
        fig.subplots_adjust(hspace=.2, wspace=.2)
        if np.max([nrows, ncols]) > 1:
            logging.debug('ravelling the matploltib as there are multipl columns')
            axs = axs.ravel(order="F")
        else:
            axs=[axs]
        if palette_choice is None:
            palette_choice = ['#1f77b4']*nrows
            logging.debug(len(palette_choice))
        # for j in range(ncols):
        idx=0
        for i, method_choice in enumerate(method_choices):
            plot_df2 = plot_df[plot_df[method] == method_choice].copy()
            for j, group_choice in enumerate(group_choices):
                logging.debug('plotting %i %s' % (j, group_choice))
                logging.debug(str(method_choice) + "|" + str(group_choice))
                scatter_one(group_choice, batch, plot_df2, axs[idx], 
                            colour=palette_choice[i], 
                            title=str(method_choice) + "|" + str(group_choice))
                idx+=1
                logging.debug(idx)
    return fig, axs
        

def facet_scatter(x, y, c, **kwargs):
    """Draw scatterplot with point colors from a faceted DataFrame columns."""
    kwargs.pop("color")
    plt.scatter(x, y, c=c, **kwargs)



def adjust_x_axis(ax):
    ax.set_xticklabels(ax.get_xticklabels(),
                       rotation= 90, 
                       horizontalalignment='center')
    ax.set(xlabel=None)


def _check_col_from_any_assay(mdata, var_col):
    if var_col in mdata.obs.columns:
        pass
    elif var_col in mdata['rna'].obs.columns: 
    # facet_row not found in obs, checking rna obs
        var_col = "rna:" + var_col
    elif var_col in mdata['prot'].obs.columns: 
    # facet_row not found in obs, checking rna obs
        var_col = "prot:" + var_col
    # finally check that it is not all NAs
    if mdata.obs[var_col].isna().sum() == mdata.shape[0]:
        logging.error("%s is all NaN, cannot plot" % var_col)
    return var_col
    

        
        

def scatter_fg_vs_bg(mdata, mdata_raw, x, y, facet_row=None):
    facet_row = _check_col_from_any_assay(mdata_raw, facet_row)
    logging.debug(facet_row)
    
    data_cols = [x for x in [x,y,facet_row] if x is not None]
    logging.debug(data_cols)
    df = pd.concat(
        [obs_df(mdata_raw, data_cols),
         obs_df(mdata, data_cols)],
        keys=["background", "cells"]).reset_index(col_level=0)
    logging.debug(df.head())
    df.rename(columns ={"level_0": "droplet_type", "level_1" : "barcode"}, inplace=True)
    g=sns.FacetGrid(data=df,  row=facet_row, col="droplet_type", height=3, aspect=1.6, margin_titles=True)
    # g.map(sns.scatterplot, x, y, s=0.2, color=".15")
    g.map(sns.histplot, x, y,cmap=sns.cubehelix_palette(start=.5, rot=-.5, as_cmap=True), bins=100 )
    ### Tidy up the labels
    # Iterate thorugh each axis
    for ax in g.axes.flat:
        # Make x and y-axis labels slightly larger
        ax.set_xlabel(ax.get_xlabel(), fontsize='large')
        ax.set_ylabel(ax.get_ylabel(), fontsize='large')

        # Make title more human-readable and larger
        if ax.get_title():
            ax.set_title(ax.get_title().split('=')[1],
                         fontsize='large')

        # Make right ylabel more human-readable and larger
        # Only the 2nd and 4th axes have something in ax.texts
        if ax.texts:
            # This contains the right ylabel text
            txt = ax.texts[0]
            ax.text(txt.get_unitless_position()[0], 
                    txt.get_unitless_position()[1],
                    txt.get_text().split('=')[1],
                    transform=ax.transAxes,
                    va='center',
                    fontsize='large'
                   )
            # Remove the original text
            ax.texts[0].remove()
    return g



from sklearn.neighbors import KernelDensity
def _kde_curve(xx, df, bandwidth=0.1):
    X=df.loc[:,xx].values[:, np.newaxis]
    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(X) 
    x=np.linspace(X.min().min(),X.max().max(),100)[:, np.newaxis]
    log_density_values=kde.score_samples(x)
    density=np.exp(log_density_values).ravel()
    return x.ravel(), density

def ridgeplot(adata, features, layer=None, splitplot=3, bandwidth=0.1):
    """
    # code based on https://github.com/rougier/scientific-visualization-book/blob/master/code/anatomy/zorder-plots.py
    """
    # get data 
    if layer is not None:
        df = obs_df(adata, keys=features, layer=layer)
    else:
        df = obs_df(adata, keys=features)

    ncols = splitplot
    nrows = int(np.ceil(len(features)/ncols))

    np.random.seed(123)
    cmap = get_cmap("Spectral")
    
    fig_width = max([5, splitplot*5])
    fig_height = max([6, int(nrows/3)])
    fig = plt.figure(figsize=(fig_width, fig_height ))

    ax = None
    for n in range(ncols):
        ax = plt.subplot(1, ncols, n + 1, frameon=False, sharex=ax)
        n_feat = len(features)
        features_sub = features[n*nrows: n*nrows+nrows]
        # print(features_sub)
        for i, feat in enumerate(features_sub):
            X, Y = _kde_curve(feat, df, bandwidth=bandwidth)
            ax.plot(X, 1.5 * Y + i, color="k", linewidth=0.75, zorder=100 - i)
            color = cmap(i / len(features_sub))
            ax.fill_between(X, 1.5 * Y + i, i, color=color, zorder=100 - i)
        # if layer == "clr":
        #     ax.set_xlim(0, 3)

        ax.yaxis.set_tick_params(tick1On=False)
        ax.set_ylim(-1, len(features_sub) + 3)

        ax.yaxis.set_tick_params(labelleft=True)
        ax.set_yticks(np.arange(len(features_sub)))
        ax.set_yticklabels(features_sub)
        for tick in ax.yaxis.get_major_ticks():
            tick.label.set_fontsize(10)
            tick.label.set_verticalalignment("bottom")

    fig.tight_layout()
    return fig, ax



def get_layer(key, mdata, layers):
    key_in_mod = {m: key in mdata.mod[m].var_names for m in mdata.mod}
    try:
        layer = [layers[k] for k,v in key_in_mod.items() if v is True][0]
        if len(layer) == 1:
            layer = layer[0]
        return layer
    except IndexError:
        print("key not found")
        return None

# https://stackoverflow.com/questions/24516340/replace-all-occurrences-of-item-in-sublists-within-list
def subst(lst, val, rep):
    result = []
    for item in lst:
        if type(item) == list:
            result.append(subst(item, val, rep))
        elif item == val:
            result.append(rep)
        else:
            result.append(item)
    return result

def plot_scatters(mdata, features_list, layers_list):
    layers_listed = [x if type(x) is list else [x] for x in layers_list]
    layers_listed =   subst(layers_listed, "X", None)
#     print(layers_listed)
    fig_dims = [len(x) for x in layers_listed]

    fig, ax = plt.subplots(nrows=fig_dims[0], ncols=fig_dims[1], figsize=(6*fig_dims[1], 6*fig_dims[0]))

    try:
        ax=ax.ravel()
    except AttributeError:
        ax=[ax]
    for ix, x in enumerate(list(itertools.product(*layers_listed))):
        print(ix, x)
        plt_title = "layers:" + " - ".join([x1 if x1 is not None else "X" for x1 in x ])
        if len(x) == 3:
            plt_title = plt_title + "\ncolor:" + features_list[2] 
        mu.pl.scatter(mdata, 
                      features_list[0],
                      features_list[1],
                      color = features_list[2], 
                      layers=x, show=False,
                      save=False, ax= ax[ix] 
                     )
        ax[ix].set_title(plt_title)
    
        # Shrink current axis by 20%
#         box = ax[ix].get_position()
#         ax[ix].set_position([box.x0, box.y0, box.width * 0.8, box.height])

#         # Put a legend to the right of the current axis
#         ax[ix].legend(loc='center left', bbox_to_anchor=(1, 0.5))
    return fig, ax





import scanpy as sc
from muon import read, MuData
from anndata import AnnData
import argparse
import pandas as pd
import numpy as np
from scipy.sparse import issparse
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.scmethods import find_all_markers_pseudo_seurat


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


parser = argparse.ArgumentParser()
parser.add_argument("--infile",
                    default="anndata_log1p.h5ad",
                    help="file name, format: .h5ad")
parser.add_argument('--modality',
                    default=None,
                    help='')
parser.add_argument('--layer',
                    default="logged_counts",
                    help='')
parser.add_argument("--output_file_prefix",
                    default="./",
                    help="output directory/markers")
parser.add_argument("--cluster_file",
                    default=None,
                    help="file name, format: .h5ad")
parser.add_argument("--testuse", default='wilcoxon', help="wilcoxon, t-test or logreg")
parser.add_argument("--pseudo_seurat", default=False, type=check_for_bool,
                    help="apply seurat like filtering before marker test")
parser.add_argument("--minpct", type=str, default='0.1', help="minimum fraction of cells expressing gene")
parser.add_argument("--mindiffpct", type=str, default='-inf',
                    help="minimum fraction difference between cluster and other cells. Setting not recommended.")
parser.add_argument("--threshuse",  type=str, default='0.25',
                    help="testing limited to genes with this (log scale) difference in mean expression level.")
parser.add_argument("--mincells", type=str, default='3',
                    help="minimum number of cells required (applies to cluster and to the other cells)")


args, opt = parser.parse_known_args()


# script -------------------------------

def check_log1p_dict(adata):
    # temp code to get around this bug
    # https://github.com/scverse/muon/issues/40
    if 'log1p' in adata.uns.keys():
        if not adata.uns['log1p']:
            L.debug( "log1p dict is Empty, refilling dict")
            adata.uns['log1p'] = {'base': None}


def load_and_merge_clusters(adata, cluster_file):
    # load clusterss
    df = pd.read_csv(cluster_file, sep="\t", index_col=0)
    adata_shape = adata.shape
    # add clusters to adata.obs
    adata.obs = pd.merge(adata.obs, df, how="left", left_index=True, right_index=True)
    # check check we have not lost cells in merge
    if adata.shape != adata_shape:
        L.info("some cells lost in merge, not all cells have a cluster?")
    L.debug(adata.shape)


def get_header():
    # write out the correct header before we start, need to werite it out before, because we use append mode for the multimodal find markers
    # and we don't wantto end up with a new header half
    if sc.__version__ < "1.7.0":
        markers_columns = ['cluster', 'scores', 'gene', 'avg_logFC', 'pvals', 'p.adj.bonferroni']
    elif sc.__version__ >= "1.7.0":
        markers_columns = ['cluster', 'gene', 'scores', 'avg_logFC', 'pvals', 'p.adj.bonferroni']
    return markers_columns


def filter_zero_count_genes(adata, layer=None):
    if layer is None:
        if issparse(adata.X):
            bool_list = (adata.X.sum(axis=0) > 0).tolist()[0]
        else:
            bool_list = (adata.X.sum(axis=0) > 0).tolist()
    else:
        if issparse(adata.layers[layer]):
            bool_list = (adata.layers[layer].sum(axis=0) > 0).tolist()[0]
        else:
            bool_list = (adata.layers[layer].sum(axis=0) > 0).tolist()
    adata = adata[:, bool_list]
    L.debug(adata.shape)


def run_clustering(adata, 
                   cluster_col=None, 
                   cluster_groups='all', 
                   layer=None, 
                   methoduse=None, 
                   pseudo_seurat=False):
    if type(cluster_groups) is list:
        # make sure the clusters are strings else rank_gene_groups crashes
        cluster_groups = [str(x) for x in cluster_groups]
    adata.obs[cluster_col] = adata.obs[cluster_col].astype('str').astype('category')
    if pseudo_seurat:
        L.info('running rank gene groups with pseudo-seurat method')
        markers, filter_stats = find_all_markers_pseudo_seurat(adata,   
                                        groups=cluster_groups, 
                                        groupby=cluster_col, 
                                        method=methoduse,
                                        n_genes=float("inf"), 
                                        corr_method="bonferroni",
                                        layer=layer,
                                        arg_minpct=0.1,
                                        arg_mindiffpct=-float("inf"), 
                                        arg_logfcdiff=0.25)
        markers = markers.reset_index().drop(columns='level_1').rename(columns={'level_0':'gene'})
    else:
        # find markers
        L.info('running rank gene groups with standard scanpy implementation')
        sc.tl.rank_genes_groups(adata, 
                                groups=cluster_groups, 
                                groupby=cluster_col, 
                                method=methoduse,
                                n_genes=float("inf"), 
                                corr_method="bonferroni",
                                layer=layer)
        markers = sc.get.rank_genes_groups_df(adata, group=None)
        L.info("Rank gene groups complete")
        filter_stats = None
    return markers, filter_stats


def main(adata, 
         cluster_file,
         mod, 
         mincells=None,
         layer=None,
         testuse=None,
         pseudo_seurat=False,
         output_file_prefix="markers") :
    # check the X slot actually contains data
    if adata.X.shape[1] == 0:
        L.info("exiting because adata.X contains no values")
        sys.exit("exiting because adata.X contains no values")
    # prevent log1p base error.
    check_log1p_dict(adata)
    # load clusters
    load_and_merge_clusters(adata, cluster_file)
    # filter out genes with zero counts
    filter_zero_count_genes(adata, layer=layer)
    # filter out clusters with less than min cells
    if mincells is not None:
        min_cell_df = adata.obs['clusters'].value_counts() 
        if any(min_cell_df < mincells):
            exclude = min_cell_df[min_cell_df < mincells].index
            L.info('exluding clusters for haveing too few cells: %s' % str(list(exclude)))
            clust_vals = list(set(min_cell_df[min_cell_df >= mincells].index))
        else:
            clust_vals = list(set(adata.obs["clusters"]))
    all_markers, all_filter_stats = run_clustering(adata,
                                                   cluster_groups=clust_vals,
                                                   cluster_col="clusters",
                                        layer=layer, 
                                        methoduse=testuse, 
                                        pseudo_seurat=pseudo_seurat)
    # make sure all files have consitent headers
    all_markers.columns = get_header()
    all_markers['mod'] = mod
    all_markers.to_csv(output_file_prefix + ".txt", index=False, sep='\t')
    # subset to significant markers
    signif_markers = all_markers[(all_markers["p.adj.bonferroni"] < 0.05) &( all_markers.avg_logFC > 0)]
    if signif_markers.shape[0] !=0:
        signif_markers.to_csv(output_file_prefix + mod + "_signif.txt", index=False, sep='\t')
        # write out to excel
        excel_file_top = output_file_prefix + mod + "_signif.xlsx"
        with pd.ExcelWriter(excel_file_top) as writer:
            for xx in signif_markers.cluster.unique():
                df_sub = signif_markers[signif_markers.cluster == xx]
                if df_sub.shape[0] !=0:
                    df_sub.to_excel(writer, sheet_name="cluster" + str(xx), index=False)
    #  when relevent write out filter stats
    if pseudo_seurat is True:
        all_filter_stats = all_filter_stats.reset_index().rename(columns={'index':'cluster'})
        all_filter_stats.to_csv(output_file_prefix + "_filter_stats_" + mod + ".txt", index=False, sep="\t")



L.info("Running with options")
L.info(args)
L.info('\n')

# read data
mdata = read(args.infile)
    

if type(mdata) is AnnData:
    adata = mdata
    # main function only does rank_gene_groups on X, so 
elif type(mdata) is MuData and args.modality is not None:
    adata = mdata[args.modality]    
else:
    sys.exit('if inputting a mudata object, you need to specify a modality')
    


main(adata, 
     mod=args.modality,
     cluster_file=args.cluster_file,
    mincells=int(args.mincells),
    layer=args.layer,
    testuse=args.testuse,
    pseudo_seurat=args.pseudo_seurat,
    output_file_prefix=args.output_file_prefix)

L.info("done")




'''
scanpy QC script PROT
order of QC:
- RNA
- PROT
- Repertoire
- ATAC
'''
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import scipy.io
import matplotlib.pyplot as plt
plt.ioff()
import numpy as np
import pandas as pd
import os
import argparse
import scanpy as sc
import seaborn as sns
import muon as mu

from panpipes.funcs.io import  write_obs
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.scmethods import identify_isotype_outliers
from panpipes.funcs.plotting import adjust_x_axis


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
# required option
parser.add_argument("--sampleprefix",
                    default="",
                    help="prefix to prepend when saving the metadata file")
parser.add_argument("--input_anndata",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--outfile",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/prot/",
                    help="path to save the figures to")
parser.add_argument('--identify_isotype_outliers',
                    default=False, type=check_for_bool,
                    help='')
parser.add_argument('--isotype_upper_quantile',
                    default=0.9,
                    help='')
parser.add_argument('--isotype_n_pass',
                    default=2,
                    help='')
parser.add_argument('--channel_col',
                    default=None,
                    help='')
parser.add_argument("--per_cell_metrics",
                    default=None,
                    help="")
parser.add_argument("--per_prot_metrics",
                    default=None,
                    help="")


args, opt = parser.parse_known_args()

L.info("Scanpy qc pipeline")
sc.settings.verbosity = 3

L.info("running with args:")
L.info(args)
figdir = args.figdir

if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))



mdata = mu.read(args.input_anndata)
prot = mdata['prot']


# determine what QC metrics we want
if args.per_cell_metrics is not None:
    per_cell_metrics = args.per_cell_metrics.split(",")
    per_cell_metrics = [a.strip() for a in per_cell_metrics]




# work out if we already have isotype column, if not try to infer from index.
if 'isotype' not in prot.var.columns:
    # this means that isotype column was not included in the protein conversion table 
    # so we are going to have a wwhack at identifying them
    prot.var['isotype'] = prot.var_names.str.contains("isotype")
else:
    # there is already an isotype column in var
    L.info("isotype column already in var")
    pass

# are there any actual isotypes though?
isotypes = list(prot.var_names[prot.var.isotype].unique())
L.info("n isotypes found in data %i" % len(isotypes))
if len(isotypes) > 0:
    qc_vars=["isotype"]
else:
    qc_vars=[]
# if isotypes are found then include them in the caluclate qc metrics call.
# this calculates standard metrics 
sc.pp.calculate_qc_metrics(prot,
                        var_type="prot",  
                        qc_vars=qc_vars, 
                        percent_top=None,log1p=True, inplace=True)

## let's assess the isotype outlier cells. 
#(Cells with an excessive amount of isotype indicating stickiness)
if (len(isotypes) > 0) & check_for_bool(args.identify_isotype_outliers):
    L.info("identifying isotype outliers")
    # this measn we found some isotypes earlier
    # we identify outliers on a per channel basis 
    # using the groupby argument
    identify_isotype_outliers(prot, isotypes, 
            quantile_val=float(args.isotype_upper_quantile),
            n_isotypes_pass=int(args.isotype_n_pass), 
            groupby=args.channel_col)

mdata.update()

# write out the cell_metadata
logging.info("saving mudata and obs in a metadata tsv file")
write_obs(mdata, output_prefix=args.sampleprefix, 
        output_suffix="_cell_metadata.tsv")
# write out whole object data.
mdata.write(args.outfile)




# now run some per channel metrics - this is not stored in the final object.
# the reason these are not run on the whole object is to get the "var_df" 
# values on a per channel basis 
# IS this cause you need the pandas per channel to plot? YES

# the rest of the script is concerned with Per PROT metrics instead of per cell. 
# per cell metrics are plotted in plotqc.R


channel_col = args.channel_col # TO DO why is this just not simply sample_id 

out = {}
for si in prot.obs[channel_col].unique():
    L.info(si)
    obs_df, var_df = sc.pp.calculate_qc_metrics(prot[prot.obs[channel_col]==si],
                            var_type="prot", 
                            qc_vars=qc_vars, 
                            percent_top=None,log1p=True, inplace=False)
    # store the var df for the metrics we care about in the dictionary.
    out[si] = var_df.copy()
    # add an extra column t the var_df to indicate which channel the data came from
    out[si][channel_col] = si

# concatenate all the var_dfs, so we can summarise the per channel data in boxplots.
var_dat = pd.concat(out).reset_index().rename(columns={'level_0':channel_col, 'level_1': "prot_id"})
prot_id_col=var_dat.columns[1]

if args.per_prot_metrics is not None:
    per_prot_metrics = args.per_prot_metrics.split(",")
    per_prot_metrics = [a.strip() for a in per_prot_metrics]

    # let's make some boxplot figures.
    for qcp in per_prot_metrics:
        fig, ax = plt.subplots(nrows=1,figsize=(24,6))
        sns.boxplot(data=var_dat, x=prot_id_col, y=qcp, ax=ax)
        adjust_x_axis(ax)
        fig.tight_layout()
        plt.savefig(os.path.join(figdir, "boxplot_" + channel_col + "_" + qcp + ".png"))
    plt.close()

    # write out the per channel metrics in a separate csv.
    var_dat.to_csv(args.sampleprefix + "_prot_qc_metrics_per_" + channel_col + ".csv")

L.info("Done")




import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import muon as mu
import os
import re
from matplotlib import cm
from matplotlib.colors import ListedColormap
import yaml 
from panpipes.funcs.io import read_yaml
from panpipes.funcs.plotting import batch_scatter_two_var
from pandas.api.types import is_string_dtype

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


parser = argparse.ArgumentParser()
parser.add_argument("--cell_meta_df")
parser.add_argument("--combined_umaps_tsv")
parser.add_argument("--fig_dir")
parser.add_argument("--batch_dict", default=None)
parser.add_argument("--qc_dict", default=None)

args, opt = parser.parse_known_args()


L.info(args)
palette_choice = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
# cmap_choice = "BuPu" # previous default = "viridis"
bupu = cm.get_cmap('BuPu', 512)
cmap_choice= ListedColormap(bupu(np.linspace(0.1, 1, 256)))


# load metadata
cell_meta_df = pd.read_csv(args.cell_meta_df, index_col=0)
umaps_df = pd.read_csv(args.combined_umaps_tsv, sep='\t', index_col=0)
umaps_df = pd.merge(umaps_df, cell_meta_df, left_index=True, right_index=True)

batch_dict = read_yaml(args.batch_dict)

# parse the qc dict
qc_dict = args.qc_dict
if isinstance(args.qc_dict, dict):
    qc_dict = args.qc_dict
else:
    qc_dict = read_yaml(args.qc_dict) 

# extract the grouping vars which are expected to be categorical
grpvar = [x.replace (" " ,"") for x in qc_dict['grouping_var'].split(",")]
grpvar = [value for value in grpvar if value in cell_meta_df.columns] 
del qc_dict['grouping_var']
L.info("additional grouping vars:")
# merge with the batch columns 
from itertools import chain
all_grpvar = list(batch_dict.values()) + [grpvar]
# unnest the list
grpvar = list(set(chain(*all_grpvar)))
L.info(grpvar)

# modality level metrics
# remove metrics from keys
qc_dict = {re.sub("_metrics", "", k) : v for k, v in qc_dict.items()}

# split the comma sep string of metrics (and remove Nones)
qc_dict = {k: v.split(',') for k, v in qc_dict.items() if v is not None}
# remove white spaces from list of the dictionary
qc_dict ={i: [a.strip(" ") for a in j ] for i,j in qc_dict.items()}
# make sure mod: prefaces each metric
qc_dict = {k: [k + ":" + vi if ":" not in vi else vi for vi in v ]  for k, v in qc_dict.items()}
L.info("additional qc vars:")
L.info(qc_dict)


# do plots
for mod in umaps_df['mod'].unique():
    # create directory if it doesn't exist
    if os.path.exists(os.path.join(args.fig_dir, mod)) is False:
        os.mkdir(os.path.join(args.fig_dir, mod))
    L.info("plotting modality: %s" % mod)
    plt_df = umaps_df[umaps_df['mod'] == mod].copy()
    pointsize = 10000 / plt_df.shape[0]
    plt_df["method"] = plt_df["method"].astype("category")
    # put none at the top of the list
    if mod != "multimodal":
        umaps_methods_list = plt_df['method'].cat.categories.tolist()
        umaps_methods_list.insert(0, umaps_methods_list.pop(umaps_methods_list.index("none")))
        plt_df['method'] = plt_df['method'].cat.reorder_categories(umaps_methods_list, ordered=True)
    #  get grouing vars
    if mod !="multimodal":
        if mod in batch_dict.keys() :
            columns = set(batch_dict[mod] + [mod +":"+ g for g in grpvar])
        else:
            # this means no batch correction was run. 
            columns = set([mod +":"+ g for g in grpvar])
    else:
        columns = grpvar
    # keep only the columns that are actually in plt_df
    columns = [cl for cl in columns if cl in plt_df.columns ]
    for col in columns:
        L.info("plotting grouping var %s" % col)
        # make sure everything is a category so that when they get plotted it's the same color
        plt_df[col] = plt_df[col].astype(str).astype('category')
        plt_df = plt_df.sample(frac=1).reset_index(drop=True)
        #plt_df = plt_df.sort_values(by="umap_1")
        g = sns.FacetGrid(plt_df, col="method", col_wrap=3, sharex=False, sharey=False)
        g = (g.map(sns.scatterplot, "umap_1", "umap_2", col, s=pointsize, linewidth=0))
        g.add_legend() 
        g.savefig(os.path.join(args.fig_dir, mod, "umap_method_" + str(col) + ".png"))
        fig, ax = batch_scatter_two_var(plt_df, "method", col, palette_choice=palette_choice)
        if fig is not None:
            fig.savefig(os.path.join(args.fig_dir, mod,  "umap_method_facet_" + str(col) + ".png"), dpi=300)
        plt.clf()

    ncats  = len(plt_df['method'].unique())
    if ncats > 4:
        ncols=4
        nrows = int(np.ceil(ncats/4))
    else:
        ncols=int(ncats)
        nrows=1

    qcmetrics = []
    if mod in qc_dict.keys() and qc_dict[mod] is not None:
        qcmetrics =  qcmetrics + qc_dict[mod] 
    if "all" in qc_dict.keys() and qc_dict["all"] is not None:    
        qcmetrics = qcmetrics + qc_dict['all'] 
    if len(qcmetrics) == 0:
        next
    # but we'll still check
    # keep only the qc metrics that are actually in plt_df
    else:
        qcmetrics = [qc for qc in qcmetrics if qc in plt_df.columns ]
        if len(qcmetrics) > 0:
            for qc in qcmetrics:
                fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=(5*ncols,4*nrows))
                if nrows > 1 or ncols > 1:
                    axes = axes.ravel()
                else:
                    axes=[axes]
                if pd.api.types.is_numeric_dtype(plt_df[qc]):
                    L.info("plotting qc var (numeric) %s" % qc)
                    for idx, mm in enumerate(plt_df['method'].cat.categories):
                        im = axes[idx].scatter(data=plt_df[plt_df['method']== mm], 
                                        x="umap_1",
                                        y="umap_2",
                                        c=qc, 
                                        s=pointsize,
                                        cmap=cmap_choice)
                        axes[idx].set_title(mm)
                        axes[idx].axis('off')
                    # delte the last few empty plots.
                    for idx2 in range(idx,ncols*nrows):
                        axes[idx2].axis('off')
                    if pd.api.types.is_numeric_dtype(plt_df[qc]):
                        fig.subplots_adjust(top=0.925,right=0.8)
                        cbar_ax = fig.add_axes([0.85, 0.35, 0.025, 0.35])
                        fig.colorbar(im, cbar_ax)
                    fig.suptitle(qc)
                    plt.savefig(os.path.join(args.fig_dir, mod , "umap_method_" + qc + ".png"), dpi = 300)
                    plt.clf()
                else:
                    # this is a categorical colored plot
                    plt_df[qc] = plt_df[qc].fillna('nan')
                    if len(plt_df[qc].unique()) < 40:
                        L.info("plotting qc var (categorical) %s"%qc)
                        g = sns.FacetGrid(plt_df, col="method", col_wrap=3, sharex=False, sharey=False)
                        g = (g.map(sns.scatterplot, "umap_1", "umap_2", qc, s=pointsize, linewidth=0))
                        g.add_legend() 
                        g.savefig(os.path.join(args.fig_dir, mod, "umap_method_" + qc + ".png"), dpi = 300)
                        plt.clf()
                    else:
                        L.info('skipping plot as too many categorys %s' % qc )

L.info('done')


import pandas as pd
import scanpy as sc
import argparse
import os
import harmonypy as hm
import muon as mu
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice

import multiprocessing 
threads_available = multiprocessing.cpu_count()

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--modality',
                    default='rna',
                    help='')
parser.add_argument('--dimred',
                    default='PCA',
                    help='which dimred to expect, relevant for ATAC')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_harmony.csv',
                    help='')
parser.add_argument('--integration_col',
                    help='')
parser.add_argument('--n_threads', default=1,
                    help="num threads to use for neighbor computations")
parser.add_argument('--harmony_npcs', default=30,
                    help="npcs for running harmony")        
parser.add_argument('--sigma_val', default=0.1,
                    help="sigma")
parser.add_argument('--theta_val', default=1.0,
                    help="theta")                   
parser.add_argument('--neighbors_n_pcs', default=30,
                    help="n_pcs")
parser.add_argument('--neighbors_method',
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k',
                    help="neighbors k")
parser.add_argument('--neighbors_metric',
                    help="neighbor metric, e.g. euclidean or cosine")


args, opt = parser.parse_known_args()


L.info("reading data and starting integration pipeline with script: ")
L.info(os.path.basename(__file__))
L.info("Running with options: %s", args)

# this should be an object that contains the full log normalised data (adata_log1p.h5ad)
# prior to hvgs and filtering
#adata = read_anndata(args.input_anndata, use_muon=use_muon, modality=args.modality)
mdata = mu.read(args.input_anndata)
adata = mdata.mod[args.modality] 

# Harmony can integrate on 2+ variables,
# but for consistency with other approaches create a fake column with combined information
columns = [x.strip() for x in args.integration_col.split(",")]
if args.dimred == "PCA":
    dimred = "X_pca"
elif args.dimred == "LSI":
    dimred = "X_lsi"

if dimred not in adata.obsm:
    L.info("i need a dimred to start, computing pca with default param")
    dimred = "X_pca" 
    n_pcs = 50
    if adata.var.shape[0] < n_pcs:
        L.info("You have less features than number of PCs you intend to calculate")
        n_pcs = adata.var.shape[0] - 1
        L.info("Setting n PCS to %i" % int(n_pcs))    
    sc.pp.scale(adata)
    sc.tl.pca(adata, n_comps=n_pcs, 
                    svd_solver='arpack', 
                    random_state=0) 



if len(columns)>1: 
    L.info("using 2 columns to integrate on more variables")
    #comb_columns = "_".join(columns)
    adata.obs["comb_columns"] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)

    # make sure that batch is a categorical
    adata.obs["comb_columns"] = adata.obs["comb_columns"].astype("category")
    # run harmony

    ho = hm.run_harmony(adata.obsm[dimred][:,0:int(args.harmony_npcs)], adata.obs, ["comb_columns"], 
                                       sigma = float(args.sigma_val),theta = float(args.theta_val),verbose=True,max_iter_kmeans=30, 
                                       max_iter_harmony=40)

else:
    # make sure that batch is a categorical
    adata.obs[args.integration_col] = adata.obs[args.integration_col].astype("category")
    # run harmony
    ho = hm.run_harmony(adata.obsm[dimred][:,0:int(args.harmony_npcs)],
                        adata.obs,
                        [args.integration_col],
                        sigma=float(args.sigma_val),
                        theta = float(args.theta_val),
                        verbose=True,max_iter_kmeans=30,
                        max_iter_harmony=40)


L.info("integration run now calculate umap")

adjusted_pcs = pd.DataFrame(ho.Z_corr).T
adata.obsm['X_harmony']=adjusted_pcs.values
L.info("harmony co-ords derived")

if int(args.neighbors_n_pcs) >adata.obsm['X_harmony'].shape[1]:
    L.warn(f"N PCs is larger than X_harmony dimensions, reducing n PCs to  {adata.obsm['X_harmony'].shape[1] -1}")

n_pcs= min(int(args.neighbors_n_pcs), adata.obsm['X_harmony'].shape[1]-1)

# run neighbours and umap 
run_neighbors_method_choice(adata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, 
    metric=args.neighbors_metric, 
    use_rep='X_harmony',
    nthreads=max([threads_available, 6]))


L.info("done n_neighbours, moving to umap")
sc.tl.umap(adata)
L.info("done umap, saving stuff")
# write out
umap = pd.DataFrame(adata.obsm['X_umap'], adata.obs.index)
umap.to_csv(args.output_csv)


#adata.write("tmp/harmony_scaled_adata_" + args.modality + ".h5ad")


outfiletmp = ("tmp/harmony_scaled_adata_" + args.modality + ".h5ad" )

L.info("saving harmony corrected adata")
write_anndata(adata, outfiletmp, use_muon=False, modality=args.modality)


L.info("Done")





#downsampling script
## written for this pipeline by Charlotte Rich-Griffin 2020-09-30

import scanpy as sc
import argparse
import pandas as pd
from muon import MuData
from anndata import AnnData
import muon as mu
import re

from panpipes.funcs.io import write_obs
from panpipes.funcs.processing import downsample_mudata


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

sc.settings.verbosity = 0  # verbosity: errors (0), warnings (1), info (2), hints (3)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_mudata',
                    default='data/mudata-n5000-filt.h5ad',
                    help='')
parser.add_argument('--output_mudata',
                    default='data/mudata-n5000-filt.h5ad',
                    help='')
parser.add_argument('--sampleprefix', default="", help="")
parser.add_argument('--downsample_value', default=None,
                    help='')
parser.add_argument('--downsample_col', default=None,
                    help='')
parser.add_argument('--intersect_mods', default=None,
                    help='comma separated string of modalities we want to intersect_obs on')                 


L.warning("Running filter_mudata")

parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()

# load data
mdata = mu.read(args.input_mudata)

if isinstance(mdata, AnnData):
    # convert anndata to muon
    mdata=MuData({'mod':mdata})

orig_obs = mdata.obs

L.info("Number of cells per sample \n%s" % mdata.obs.sample_id.value_counts())


if args.downsample_col == "None":
    args.downsample_col = None

if args.intersect_mods is None:
    # we assume all mods
    mods = list(mdata.mod.keys())
else:
    mods = args.intersect_mods.split(',')


# remove unused categories in batch col
if args.downsample_col is not None:
    mdata.obs[args.downsample_col] = mdata.obs[args.downsample_col].cat.remove_unused_categories()

# do the downsample.

downsample_mudata(mdata, nn=int(args.downsample_value),
    cat_col=args.downsample_col,
    mods = mods,
    inplace=True)
print(mdata)


mdata.update()
    
L.info("Number of cells per sample")
L.info(mdata.obs.sample_id.value_counts())

# write out the observations
write_obs(mdata, output_prefix=re.sub("\\.(.*)", "", args.output_mudata), 
        output_suffix="_downsampled_cell_metadata.tsv")


mdata.write(args.output_mudata)
#This stage is the point (i.e. pre-normalisation) where the mdata file can be outputted so that we can extract raw matrix for the cellphonedb.
L.info("Done")


"""
Plotting umaps of a small list of custom genes
CRG 2020-06-19
"""
import scanpy as sc
import muon as mu
from anndata import AnnData
import pandas as pd
import argparse
from panpipes.funcs.io import read_yaml
from itertools import product

sc.settings.autoshow = False
import os
import matplotlib
matplotlib.use('agg')

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


# parse argumetns
parser = argparse.ArgumentParser()
parser.add_argument("--infile",
                    default="pbmc3k.h5mu",
                    help="file name, format: .h5mu/.h5ad")
parser.add_argument("--modalities",
                    default="rna",
                    help="modality keys containing data to be plotted, comma spearated list e.g. rna,prot default 'rna'")
parser.add_argument("--layers",
                    default="X",
                    help="layer containing data to be plotted, default=X")
parser.add_argument("--basis_dict",
                    default="X_umap",
                    help="basis in obsm to plot")
parser.add_argument("--marker_file",
                    default="Marker_Lists-myeloid.csv",
                    help="comma separateed list of marker_list csvs, one column of gene_ids")
parser.add_argument("--base_figure_dir", default="figures/",
                    help="figures path")

args, opt = parser.parse_known_args()
L.info("running with:")
L.info(args)
sc.settings.figdir = args.base_figure_dir

# ---- script

def main(adata, mod, layer_choice, df, basis):
    for gc in df['group'].unique():
        # define file name
        if layer_choice is None or layer_choice =="X":
            layer_choice = None
            layer_string = ""
        else:
            layer_string = layer_choice
        fname_prefix = "_".join(["_" + mod, layer_string, gc])
        # get features
        fetches = df[df['group'] == gc]['feature']
        plot_features = [gg for gg in fetches if gg in adata.var_names]
        plot_features = list(set(plot_features))
        L.info("plotting")
        L.info(plot_features)
        sc.settings.figdir  = os.path.join(args.base_figure_dir, mod)
        mu.pl.embedding(adata, basis=basis, layer=layer_choice, color=plot_features, save = fname_prefix + ".png")


L.debug("load data")
mdata = mu.read(args.infile)
modalities = args.modalities.split(',')

L.debug("load marker file")
df = pd.read_csv(args.marker_file )

# get layer
try:
    layers = read_yaml(args.layers)
except AttributeError:
    # this assumes that we have tried to parse a dict and nstead found a string
    # there is probably a better solution
    layers = args.layers

# get bases
try:
    basis_dict = read_yaml(args.basis_dict)
except AttributeError:
    # this assumes that we have tried to parse a dict and nstead found a string
    # there is probably a better solution
    basis_dict = args.basis_dict


if type(mdata) is AnnData:
    adata = mdata
    basis= basis_dict
    main(adata, layer_choice=args.layers,  df = df, basis=basis)
else:
    # we have multimodal object
    for mod in modalities:
        print(mod)
        df_sub = df[df['mod'] == mod]
        mdata.update_obs()
        try:
            ll = layers[mod]
        except KeyError:
            ll = [None]
        if mod in basis_dict.keys():
            bb= basis_dict[mod]
        else:
            bb = []
        if len(bb) > 0 :
            for basis, layer in product(bb, ll):
                print(basis,layer)
                main(adata=mdata[mod], 
                    mod=mod,
                    layer_choice = layer,
                    df=df_sub,
                    basis=basis)

L.info("Done")




import argparse
import numpy as np
import pandas as pd
import muon as mu
import os
import re
import sys
import logging
import yaml

L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


parser = argparse.ArgumentParser()
parser.add_argument("--input_mudata")
parser.add_argument("--input_umap_files")
parser.add_argument("--output_combined_umaps_tsv")
args, opt = parser.parse_known_args()
L.info(args)
L.info("reading in files")

# cell_meta_df = mu.read(args.input_mudata+'/rna').obs
# mtd_columns = cell_meta_df.columns.to_list()



L.info("reading in all umaps")
# load files
input_files = args.input_umap_files.split(",")
umaps_list = [pd.read_csv(x, index_col=0) for x in input_files]
umaps_index = [re.sub("umap_|.csv", "", os.path.basename(x)) for x in input_files]
umaps_df = pd.concat(umaps_list, axis=0, keys=umaps_index).reset_index(level=0)
umaps_df.columns = ["method", "umap_1", "umap_2"]
# umaps_df = umaps_df.merge(cell_meta_df, left_index=True, right_index=True)

# reorder methods so None is at the most left on plot
umaps_df['method'] = [x.split('_')[0] for x in umaps_df.method]

umaps_df = umaps_df.sort_values(by='method')

# save umap df to file
L.info("save umap df to file")
umaps_df.iloc[:,0:4].to_csv(args.output_combined_umaps_tsv, sep="\t")

L.info("done")




import multiprocessing
from matplotlib import transforms 

import anndata as ad
import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import scvi
import argparse
import os
import muon as mu
from mudata import MuData
import numpy as np
import panpipes.funcs as pp
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_yaml
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--query_data',
                    default='adata_scaled.h5ad',
                    help='path to query data. can be a raw 10x dataset or a preprocessed anndata/mudata')
parser.add_argument('--query_celltype',default=None,
                    help='if query has already a column with cell labeling in obs. Default to "celltype" ')
parser.add_argument('--adata_reference',
                    default='adata_.h5ad',
                    help='path to reference ann/mudata. necessary only if you want to produce ref/ query map')
parser.add_argument('--reference_path',
                    default='reference model directory',
                    help='')
parser.add_argument('--reference_architecture', default='scanvi',
                    help='scvi, scanvi, totalvi supported')
# parser.add_argument('--generate_scanvi', default=False,
#                     help='create scanvi reference from scvi to allow for label transfer')
parser.add_argument('--predict_rf',default=False,
                    help='For totalvi and scvi, if reference model has random forest classifier, classify cells in query')
parser.add_argument('--impute_proteins', default=False,
                    help='if totalvi reference is used, allow to impute protein levels')
parser.add_argument('--transform_batch', default=None,
                    help='if totalvi reference is used, batch categories for the query dataset')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs", default=50)
parser.add_argument('--neighbors_method', default="scanpy",
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k', default=30,
                    help="neighbors k")
parser.add_argument('--neighbors_metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")
parser.add_argument('--outfile',default=None,
                    help="file where to save umap")

args, opt = parser.parse_known_args()
sc.settings.figdir = "figures/"
sc.set_figure_params(figsize=(8, 6), dpi=300) 
L.info("running with args:")
args.predict_rf = check_for_bool(args.predict_rf)
args.impute_proteins = check_for_bool(args.impute_proteins)
L.info(args)

threads_available = multiprocessing.cpu_count()

# load parameters
params = read_yaml("pipeline.yml")
query_data = os.path.basename(args.query_data)

reference_architecture = str(args.reference_architecture)



mdata = mu.read(args.query_data)
if type(mdata) is mu.MuData:
    if "rna" not in mdata.mod.keys():
        sys.exit("we only support querying using RNA but your mdata doesn't contain rna")
    else:
        adata_query = mdata["rna"].copy()
        if "prot" in mdata.mod.keys() and reference_architecture=="totalvi":
            if X_is_raw(mdata['prot']):
                X_array = mdata['prot'].X.copy()
            elif "raw_counts" in mdata['prot'].layers.keys(): 
                X_array = mdata['prot'].layers['raw_counts'].copy()
        
        X_df = pd.DataFrame(X_array.todense(), index=mdata['prot'].obs_names, columns=mdata['prot'].var_names)
        if X_df.shape[0] == adata_query.X.shape[0]:
            L.info("adding protein_expression to obsm")
            #check the obs are in the correct order
            X_df = X_df.loc[adata_query.obs_names,:]
            adata_query.obsm['protein_expression'] = X_df 
        else:
            L.error("dimensions do not match, cannot create the raw obsm counts in query")
else:
    adata_query = mdata.copy()

L.info("this is your query anndata")
print(adata_query)

if "counts" not in adata_query.layers.keys():
    if "raw_counts" in adata_query.layers.keys():
        adata_query.layers["counts"] = adata_query.layers["raw_counts"].copy()
    elif X_is_raw(adata_query):
        adata_query.layers["counts"] = adata_query.X.copy()
    else:
        raise AttributeError("raw counts not found in query, \
                        store in either X or in 'counts/raw_counts' layer")


if reference_architecture in ['totalvi','scvi','scanvi']:
    L.info("running using %s" % reference_architecture)
else:
    sys.exit(" i don't recognise this architecture : %s") %reference_architecture

reference_path = args.reference_path
if not os.path.exists(reference_path):
    L.info("The reference path you provided does not exist")
    sys.exit("The reference path you provided does not exist")
else:
    reference_path = os.path.dirname(reference_path)
    L.debug("Reference path: %s" % reference_path)

train_kwargs = {}
if params["training_plan"][reference_architecture] is not None:
    train_kwargs = params["training_plan"][reference_architecture]
    if train_kwargs["max_epochs"] is not None:
        max_epochs = train_kwargs["max_epochs"]
        del train_kwargs["max_epochs"]
    else:
        max_epochs = 200
else:
    max_epochs = 200
    train_kwargs = {'weight_decay': 0.0}


if reference_architecture=="scvi":
    scvi.model.SCVI.prepare_query_anndata(adata_query, reference_path)
    vae_q = scvi.model.SCVI.load_query_data(
    adata_query,
    reference_path)
    latent_choice= "X_scvi"
    vae_q.train(max_epochs= max_epochs , plan_kwargs=train_kwargs)
    adata_query.obsm["X_scvi"] = vae_q.get_latent_representation()
    

if reference_architecture=="scanvi":
    # Notice that adata_query.obs["labels_scanvi"] does not exist. 
    # The load_query_data method detects this and fills it in adata_query with the unlabeled category (here "Unknown").
    scvi.model.SCANVI.prepare_query_anndata(adata_query, reference_path)
    vae_q = scvi.model.SCANVI.load_query_data( 
    adata_query,
    reference_path)
    latent_choice= "X_scanvi"
    #vae_q.train(**train_kwargs) this doesn't work anymore cause max_epochs is not recognised as part of the plan kwargs
    vae_q.train(max_epochs= max_epochs , plan_kwargs=train_kwargs)
    adata_query.obsm["X_scanvi"] = vae_q.get_latent_representation()
    adata_query.obs["predictions"] = vae_q.predict()
    if args.query_celltype is not None:
        L.info("Query has celltypes in column %s, i will plot what predictions look like from scanvi model" % args.query_celltype)
        df = adata_query.obs.groupby([str(args.query_celltype), "predictions"]).size().unstack(fill_value=0)
        norm_df = df / df.sum(axis=0)

        plt.figure(figsize=(8, 8))
        _ = plt.pcolor(norm_df)
        _ = plt.xticks(np.arange(0.5, len(df.columns), 1), df.columns, rotation=90)
        _ = plt.yticks(np.arange(0.5, len(df.index), 1), df.index)
        plt.xlabel("Predicted")
        plt.ylabel("Observed")
        plt.tight_layout()
        file_name = "SCANVI_predicted_vs_observed_labels_query_data"
        plt.savefig(os.path.join("figures/", file_name + ".png"))
        
if reference_architecture=="totalvi":
    # temporary fix is disabled for now, need to modify to allow to fix query to match reference
    fix_query = False
    if fix_query:
        L.info(" will do some manipulation of query data to make it match to referece structure")
        if args.adata_reference is not None:
            reference_data = os.path.basename(args.adata_reference)
            mdata=mu.read(args.adata_reference)
            if type(mdata) is MuData:
                if "rna" not in mdata.mod.keys():
                    sys.exit("we only support querying using RNA but your mdata doesn't contain rna")
                else:
                    adata_ref = mdata["rna"].copy()
                    adata_ref.obsm = mdata.obsm.copy()
                    adata_ref.obsp = mdata.obsp.copy()
            else:
                adata_ref = mdata.copy()
            del mdata
            # match query and reference vars
            
            #adata_query.layers["counts"] = adata_query.X.copy()#already taken care of 
            # are the following necessary?
            sc.pp.normalize_total(adata_query, target_sum=1e4)
            sc.pp.log1p(adata_query)
            adata_query.raw = adata_query
            # subset to reference vars
            adata_query = adata_query[:, adata_ref.var_names].copy()
            
            adata_query.obs["celltype.l2"] = "Unknown"
            adata_query.obs["orig.ident"] = adata_query.obs["set"]
            #query.obsm["X_umap"] = query.obs[["UMAP1", "UMAP2"]]


            # obsm slot slot for protein counts can be called either prot_counts or protein_expression
            # in the reference anndata and model. allow a 

            if "protein_expression" in adata_ref.obsm.keys():
                pname = "protein_expression"
            
            elif "protein_counts" in adata_ref.obsm.keys():
                pname = "protein_counts"
            
            if pname not in adata_query.obsm.keys():
                X_df = pd.DataFrame(0, index=adata_ref.obs_names, columns=adata_ref.obsm[pname].columns)

                adata_query.obsm["protein_counts"] = adata_query.obsm["pro_exp"].copy()
            

            for p in adata_ref.obsm[pname].columns:
                if p not in adata_query.obsm[pname].columns:
                    adata_query.obsm[pname][p] = 0.0
            # ensure columns are in same order
            adata_query.obsm[pname] = adata_query.obsm[pname].loc[:, adata_ref.obsm[pname].columns]

        else:
            
            # For now make this only work when reference anndata is available

            # if args.reference_var is not None:
            #     reference_var = pd.read(args.reference_var, sep="\t")
            # else: 
            #     sys.exit("I need var names from reference to make sure query has overlap")    
            # if args.reference_prot is not None:
            #     reference_prot = pd.read(args.reference_prot, sep="\t")
            # else: 
            #     sys.exit("I need var names from reference to make sure query has overlap")    

            # if args.reference_prot_assay is not None:
            #     L.info("name of obsm slot for reference is: %i" % args.reference_prot_assay)
            #     pname = args.reference_prot_assay
            sys.exit("need a reference dataset to check for matching query entries")        
        
    scvi.model.TOTALVI.prepare_query_anndata(adata_query, reference_path)
    vae_q = scvi.model.TOTALVI.load_query_data(
    adata_query,
    reference_path,
    freeze_expression=True)
    latent_choice= "X_totalvi"
    vae_q.train(max_epochs= max_epochs , plan_kwargs=train_kwargs)
    adata_query.obsm["X_totalvi"] = vae_q.get_latent_representation()
    #remove this after finishing
    adata_query.write(os.path.join( "query_temp_check_totvi.h5ad"))
        
    if args.predict_rf :
        L.info("predicting celltypes")
        predictions = (
           vae_q.latent_space_classifer_.predict(
            adata_query.obsm["X_totalvi"]
            )
        )
        adata_query.obs["predictions"] = predictions
        #remove this after finishing
        adata_query.write(os.path.join( "query_temp_check_predictions_totvi.h5ad"))
    
        if args.query_celltype is not None:
            L.info("Query has celltypes in column %s, i will plot what predictions look like from totalvi model" % args.query_celltype)
            df = adata_query.obs.groupby([str(args.query_celltype), "predictions"]).size().unstack(fill_value=0)
            norm_df = df / df.sum(axis=0)

            plt.figure(figsize=(8, 8))
            _ = plt.pcolor(norm_df)
            _ = plt.xticks(np.arange(0.5, len(df.columns), 1), df.columns, rotation=90)
            _ = plt.yticks(np.arange(0.5, len(df.index), 1), df.index)
            plt.xlabel("Predicted")
            plt.ylabel("Observed")
            plt.tight_layout()
            file_name = "totalvi_predicted_vs_observed_labels_query_data"
            plt.savefig(os.path.join("figures/", file_name + ".png"))
        
    
L.info("Mapped Query to Reference, now some utilities")


if args.adata_reference is not None:
    reference_data = os.path.basename(args.adata_reference)
    mdata=mu.read(args.adata_reference)
    if type(mdata) is MuData:
        if "rna" not in mdata.mod.keys():
            sys.exit("we only support querying using RNA but your mdata doesn't contain rna")
        else:
            adata_ref = mdata["rna"].copy()
            adata_ref.obsm = mdata.obsm.copy()
            adata_ref.obsp = mdata.obsp.copy()
    else:
        adata_ref = mdata.copy()

    adata_ref.obs.loc[:, 'is_reference'] = 'Reference'
    adata_query.obs.loc[:, 'is_reference'] = 'Query'
    #expect the batch to be always encoded as `batch` in both Q and R
    adata_full = ad.concat( [adata_ref,adata_query])
    # add param to decide if to recompute the total embedding or to recalc the embedding
    #if "X_totalvi" not in adata_ref.obsm.keys():
    adata_full.obsm[latent_choice] = vae_q.get_latent_representation(adata_full)
    
    if args.impute_proteins: 
        if args.transform_batch is not None:
            transform_batch = str(args.transform_batch)
        else:
            transform_batch = None
        normX, protein = vae_q.get_normalized_expression(
            adata_full,
            n_samples=25,
            return_mean=True,
            transform_batch= transform_batch) 
        adata_full.obsm["totalvi_denoised_rna"], adata_full.obsm["totalvi_denoised_protein"] = normX, protein    

    if reference_architecture=="scanvi":
        full_predictions = vae_q.predict(adata_full)
        L.warn("Acc: {}".format(np.mean(full_predictions == adata_full.obs.celltype)))
        adata_full.obs["predictions"] = full_predictions
else:
    adata_query.obs.loc[:, 'is_reference'] = 'Query'
    adata_full = adata_query.copy()

if int(args.neighbors_n_pcs) > adata_full.obsm[latent_choice].shape[1]:
    L.warn("N PCs is larger than %i dimensions, reducing n PCs to " % adata_full.obsm[latent_choice].shape[1])


n_pcs= min(int(args.neighbors_n_pcs), adata_full.obsm[latent_choice].shape[1])


run_neighbors_method_choice(adata_full, 
            method=args.neighbors_method, 
            n_neighbors=int(args.neighbors_k), 
            n_pcs=n_pcs, 
            metric=args.neighbors_metric, 
            use_rep=latent_choice,
            nthreads=max([threads_available, 6]))
sc.tl.umap(adata_full, min_dist=0.4)
sc.tl.leiden(adata_full, key_added="leiden_" + latent_choice)

model_name = os.path.basename(os.path.dirname(args.reference_path))
model_name = ''.join(e for e in model_name if e.isalnum())


file_name= "umap_" + model_name + "_" + latent_choice
L.info ("filename is %s" % file_name )

fig = sc.pl.embedding(adata_full, basis = "umap",color=["is_reference"],
         show=False, return_fig=True)
fig.tight_layout()
fig.savefig(os.path.join("figures/", file_name + ".png"))

umap = pd.DataFrame(adata_full.obsm['X_umap'], adata_full.obs.index)

umap.to_csv(os.path.join("refmap/", file_name + ".csv") )
file_name= "query_to_reference_" + model_name + "_" + latent_choice + ".h5mu" 

mdata_save = MuData({"rna":adata_full})

mdata_save.write(os.path.join( "refmap/" , file_name))

L.info('done')






import multiprocessing 

import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import argparse
import muon as mu
from cgatcore import pipeline as P
import panpipes.funcs as pp
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--scaled_anndata',
                    default='adata_scaled.h5ad',
                    help='a preprocessed mudata/anndata')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_wnn.csv',
                    help='')
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--n_neighbors',
                    help="number of neighbors", default=50)
parser.add_argument('--n_bandwidth_neighbors', default="",
                    help="")
parser.add_argument('--n_multineighbors', default=30,
                    help="neighbors k")
parser.add_argument('--metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")
parser.add_argument('--low_memory',default=True,
                    help="set to True by default if cells in dataset >50k")


args, opt = parser.parse_known_args()

L.info(args)
# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir

params = pp.io.read_yaml("pipeline.yml")
threads_available = multiprocessing.cpu_count()


if params['multimodal']['WNN']['modalities'] is not None:
    modalities= params['multimodal']['WNN']['modalities']
    modalities = [x.strip() for x in modalities.split(",")]
    L.info(f"using modalities :{modalities}")

L.info("running with batch corrections:")
wnn_params_bc = params['multimodal']['WNN']['batch_corrected'] 
if modalities is not None:
    wnn_params_bc= {k: wnn_params_bc[k] for k in wnn_params_bc.keys() & modalities}
L.info( wnn_params_bc )

mdata = mu.read(args.scaled_anndata)

if all(x in modalities for x in mdata.mod.keys()):
    tmp = mdata.copy()
    removed_mods = None
else:
    tmp = mdata.copy()
    removed_mods = list(set(mdata.mod.keys()) - set(modalities))
    L.info(f"removing modalities {removed_mods}")
    for rmod in removed_mods:
        del tmp.mod[rmod]

L.info("intersecting modality obs before running wnn")
mu.pp.intersect_obs(tmp)


#one could also check of obsp is not empty and use precomputed connectivities but i assume that if batch corrected the obsp is populated, if not, calc on the flight on pca
dict_graph = {}
for x in wnn_params_bc.keys():
    dict_graph[x] = {}
    if wnn_params_bc[x] is not None:
        dict_graph[x]["obsm"] = "X_" + wnn_params_bc[x]
        dict_graph[x]["anndata"] = "tmp/" + wnn_params_bc[x].lower() + "_scaled_adata_" + x +".h5ad"
    else: 
        dict_graph[x]["obsm"] = None

L.debug(dict_graph)

if dict_graph["rna"]["obsm"] == "X_scvi":
    dict_graph["rna"]["obsm"] = "X_scVI"

for kmod in dict_graph.keys():
    L.info(kmod)
    pkmod=params['multimodal']['WNN']['knn'][kmod]
    if dict_graph[kmod]["obsm"] is not None:
        if dict_graph[kmod]["obsm"] not in tmp.mod[kmod].obsm.keys():
            L.info("provided mdata doesn't have the desired obsm, just checking if it's bbknn you want.")
            if dict_graph[kmod]["obsm"] == "X_bbknn":
                if len(tmp.mod[kmod].obsp.keys()) > 0 and "neighbors" in tmp.mod[kmod].uns.keys() : 
                     L.info("i found a populated obsp slot and I assume it's bbknn")
                else:
                    if dict_graph[kmod]["anndata"] is not None:
                        L.info("reading precomputed connectivities for bbknn")
                        adata = mu.read(dict_graph[kmod]["anndata"])
                        tmp.mod[kmod].obsp = adata.obsp.copy()
                        tmp.mod[kmod].obsm[dict_graph[kmod]["obsm"]] = adata.obsm[dict_graph[kmod]["obsm"]].copy()
                        tmp.mod[kmod].uns["neighbors"]= adata.uns["neighbors"].copy()
                        tmp.update()
            else:
                if dict_graph[kmod]["anndata"] is not None:
                    L.info("provided mdata doesn't have the desired obsm. reading the batch corrected data from another stored object")
                    adata = mu.read(dict_graph[kmod]["anndata"])
                    L.debug(kmod + "object")
                    L.debug(adata)
                    tmp.mod[kmod].obsm[dict_graph[kmod]["obsm"]] = adata.obsm[dict_graph[kmod]["obsm"]].copy()
                    tmp.mod[kmod].obsp = adata.obsp.copy()
                    tmp.mod[kmod].uns['neighbors'] = adata.uns['neighbors'].copy()
                    tmp.update()
                    repuse = dict_graph[kmod]["obsm"] 
                else:
                    L.info("could not find the desired obsm and the anndata slot is empty, will calculate on the flight")
                    if kmod =="atac":
                        if "X_lsi" in tmp.mod[kmod].obsm.keys():
                            repuse = "X_lsi"
                        else:
                            repuse = "X_pca"
                        L.info("falling back on %s" %(repuse) )
            
                    L.info("calculating neighbours")
                    if repuse != "X_bbknn":
                        run_neighbors_method_choice(tmp.mod[kmod], 
                            method=pkmod['method'], 
                            n_neighbors=int(pkmod['k']), 
                            n_pcs=min(int(pkmod['npcs']), mdata.var.shape[0]-1), #this should be the # rows of var, not obs
                            metric=pkmod['metric'], 
                            #does this throw an error if no PCA for any single mod is stored?
                            use_rep=repuse,
                            nthreads=max([threads_available, 6]))
        else:
            L.info("Using %s" %(dict_graph[kmod]["obsm"]))            
    else:
        L.info("could not find the desired obsm and the anndata slot is empty, will calculate on the flight")
        repuse ="X_pca"
        if kmod =="atac":
            if "X_lsi" in tmp.mod[kmod].obsm.keys():
                repuse = "X_lsi"
            else:
                repuse = "X_pca"
            L.info("falling back on %s" %(repuse) )

        L.info("calculating neighbours")
        
        run_neighbors_method_choice(tmp.mod[kmod], 
            method=pkmod['method'], 
            n_neighbors=int(pkmod['k']), 
            n_pcs=min(int(pkmod['npcs']), mdata.var.shape[0]-1), #this should be the # rows of var, not obs
            metric=pkmod['metric'], 
            #does this throw an error if no PCA for any single mod is stored?
            use_rep=repuse,
            nthreads=max([threads_available, 6]))

L.debug(tmp)
tmp.update()
L.debug(tmp)
L.info("Now running WNN")

mu.pp.neighbors(tmp, 
                n_neighbors= int(args.n_neighbors),
                n_bandwidth_neighbors= int(args.n_bandwidth_neighbors),
                n_multineighbors= int(args.n_multineighbors),
                metric= args.metric,
                low_memory= check_for_bool(args.low_memory),   
                key_added='wnn')
L.info("WNN finished now calculate umap")

mu.tl.umap(tmp,min_dist=0.4, neighbors_key='wnn')
#For taking use of mdata.obsp['connectivities'], it’s scanpy.tl.leiden() that should be used. not muon.tl.leiden
sc.tl.leiden(tmp, neighbors_key='wnn', key_added='leiden_wnn') 

umap = pd.DataFrame(tmp.obsm['X_umap'], tmp.obs.index)
umap.to_csv(args.output_csv)

#add back modalities that were not used for this run

if removed_mods is not None:
    for rmd in removed_mods:
        tmp.mod[rmd] = mdata.mod[rmd].copy()

tmp.write("tmp/wnn_scaled_adata.h5mu")

L.info("Done")




'''
scanpy QC script RNA
order of QC:
- RNA
- PROT
- Repertoire
- ATAC
'''
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import argparse
import pandas as pd
import muon as mu
import scirpy as ir
import matplotlib.pyplot as plt
plt.ioff()
from panpipes.funcs.io import write_obs
import yaml

parser = argparse.ArgumentParser()
# required option
parser.add_argument("--sampleprefix",
                    default="",
                    help="prefix to prepend when saving the metadata file")
parser.add_argument("--input_mudata",
                    default="mdata_unfilt.h5mu",
                    help="")
parser.add_argument("--output_mudata",
                    default="mdata_unfilt.h5mu",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/",
                    help="path to save the figures to")

parser.add_argument("--distance_metrics",
                    default=None, type=yaml.safe_load,
                    help="what metrics to calculate sequence distance metric? \
                    any arguments from scirpy.pp.ir_dist as a dict in format'{arg: value}' ")
parser.add_argument("--clonotype_metrics",
                    default=None,type=yaml.safe_load,
                    help="what metrics to calculate sequence distance metric? \
                    any arguments from scirpy.pp.define_clonotypes as a dict in format'{arg: value}' ")
args, opt = parser.parse_known_args()
L.info(args)
mdata = mu.read(args.input_mudata)
rep = mdata['rep']
# chain qc
ir.tl.chain_qc(rep)

# remove nones, so defaults are used
if args.distance_metrics is not None:
    # remove nones, so defaults are used
    dist_args = {k: v for k, v in args.distance_metrics.items() if v != "None"}
    dist_args = {k: v for k, v in dist_args.items() if v}
else:
    dist_args = {}

L.info("distance args:")
L.info(dist_args)

ir.pp.ir_dist(rep, **dist_args)

# pull function arguments from args. 
if args.clonotype_metrics is not None:
    # remove nones, so defaults are used
    clonotype_args = {k: v for k, v in args.clonotype_metrics.items() if v != "None"}
    clonotype_args = {k: v for k, v in clonotype_args.items() if v}
else:
    clonotype_args={}
# define clonotypes
L.info("clonotypes args:")
L.info(clonotype_args)

ir.tl.define_clonotypes(rep, **clonotype_args)
ir.tl.clonal_expansion(rep)

    
category_cols = ['has_ir']
for cc in category_cols:
    if pd.api.types.infer_dtype(rep.obs[cc]) != "categorical":
        rep.obs[cc] = rep.obs[cc].astype('string').astype('category')


if "BCR" in rep.obs['receptor_type'].values:
    bcr = rep[rep.obs.receptor_type =="BCR", :].copy()
else:
    bcr=None
    
if "TCR" in rep.obs.receptor_type.values:
    tcr = rep[rep.obs.receptor_type =="TCR", :].copy()
else:
    tcr=None

# plot group abundances
fig, ax = plt.subplots()
ax = ir.pl.group_abundance(rep, groupby="receptor_type", ax=ax)
fig.savefig(args.figdir +"/barplot_group_abundance_receptor_type.png", bbox_inches="tight")

if bcr is not None:
    fig, ax = plt.subplots()
    ax = ir.pl.group_abundance(bcr, groupby="chain_pairing", ax=ax)
    fig.tight_layout()
    fig.savefig(args.figdir +"/barplot_group_abundance_bcr_receptor_subtype.png", bbox_inches="tight")

if tcr is not None:
    fig, ax = plt.subplots()
    ax = ir.pl.group_abundance(tcr, groupby="chain_pairing",  ax=ax)
    fig.tight_layout()
    fig.savefig(args.figdir +"/barplot_group_abundance_tcr_receptor_subtype.png", bbox_inches="tight")

# clonal expansion

clone_counts = rep.obs.groupby("receptor_type").clone_id.value_counts().to_frame("cell_counts").reset_index()

if bcr is not None:
    ax = ir.pl.clonal_expansion(bcr, groupby="sample_id")
    ax.set_title("bcr clonal expansion")
    plt.savefig(args.figdir +"/barplot_clonal_expansion_bcr.png", bbox_inches="tight" )

if tcr is not None:
    ax = ir.pl.clonal_expansion(tcr, groupby="sample_id")
    ax.set_title("tcr clonal expansion")
    plt.savefig(args.figdir +"/barplot_clonal_expansion_tcr.png", bbox_inches="tight")


L.info("saving anndata and obs in a metadata tsv file")
mdata.update()
write_obs(mdata, output_prefix=args.sampleprefix, 
    output_suffix="_cell_metadata.tsv")
mdata.write(args.output_mudata)

L.info("Done")




import argparse
import yaml
# import scanpy as sc
import pandas as pd
# import numpy as np
# from scipy.sparse import csr_matrix
import muon as mu
import warnings
from muon._atac.tools import add_peak_annotation, locate_fragments


import sys
import logging

L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

from panpipes.funcs.io import check_filetype, load_mdata_from_multiple_files, update_intersecting_feature_names, read_yaml
from panpipes.funcs.processing import check_for_bool, add_var_mtd, update_var_index, intersection, intersect_obs_by_mod


parser = argparse.ArgumentParser()
#parser.add_argument('--mode_dictionary',default=mode_dictionary,type=json.loads)
parser.add_argument('--mode_dictionary',default="",
                    help="accepts a yaml speficication of modalities dictionary", 
                    type=yaml.safe_load)
parser.add_argument('--sample_id',
                    default=None,
                    help='')
parser.add_argument('--rna_infile',
                    default=None,
                    help='')
parser.add_argument('--rna_filetype',
                    default=None,
                    help='')
parser.add_argument('--prot_infile', 
                    default=None,
                    help='')
parser.add_argument('--prot_filetype', 
                    default=None,
                    help='')
parser.add_argument('--subset_prot_barcodes_to_rna', 
                    default=True,
                    help='')
parser.add_argument('--atac_infile', 
                    default=None,
                    help='')
parser.add_argument('--atac_filetype', 
                    default=None,
                    help='')
parser.add_argument('--tcr_filtered_contigs', 
                    default=None,
                    help='')
parser.add_argument('--tcr_filetype', 
                    default=None,
                    help='')
parser.add_argument('--bcr_filtered_contigs', 
                    default=None,
                    help='')
parser.add_argument('--bcr_filetype',
 default=None,
                    help='')
parser.add_argument('--output_file',
                    default=None,
                    help='')
parser.add_argument('--per_barcode_metrics_file',
                    default=None,
                    help='ATAC/Multiome specific input file from csv')                    
parser.add_argument('--peak_annotation_file',
                    default=None,
                    help='ATAC/Multiome specific input file from csv')                    
parser.add_argument('--fragments_file',
                    default=None,
                    help='ATAC/Multiome specific input file from csv')                    


parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()
L.info(args)
# Make mdata object

# unimodal mu (check if all the modalities)
if isinstance(args.mode_dictionary, dict):
    mode_dictionary = args.mode_dictionary
else:
    mode_dictionary = read_yaml(args.mode_dictionary) 

permf = [key for key, value in mode_dictionary.items() if value == True]
all_files = {"rna": [args.rna_infile, args.rna_filetype], 
            "prot": [args.prot_infile, args.prot_filetype], 
            "atac": [args.atac_infile, args.atac_filetype], 
            "tcr":[args.tcr_filtered_contigs, args.tcr_filetype],
            "bcr":[args.bcr_filtered_contigs, args.bcr_filetype]}
#subset to the modalities we want from permf
all_files = {nm: x  for (nm, x) in all_files.items() if nm in permf}
# this will work for any combination of modalities or number of modalities in all_files.
[check_filetype(x[0], x[1]) for x in all_files.values()]
mdata = load_mdata_from_multiple_files(all_files)
                
# now lets do some extra processing on the different modalities
if 'rna' in mdata.mod.keys():
    mdata['rna'].var_names_make_unique()
    # keep only rna with at least 1 count (relevent when loading raw data)
    mdata['rna'].obs['total_counts'] = mdata['rna'].X.sum(axis=1) 
    mu.pp.filter_obs(mdata['rna'], 'total_counts', lambda x: (x > 1))
    # drop this columns as we don't need it anymore, and it will confuse things downstream
    mdata['rna'].obs = mdata['rna'].obs.drop(columns="total_counts")
    mdata.update()

if 'prot' in mdata.mod.keys():
    if 'rna' in mdata.mod.keys():
        if check_for_bool(args.subset_prot_barcodes_to_rna):
            intersect_obs_by_mod(mdata, ['rna', 'prot'])
            mdata.update()
    L.info(mdata['prot'].var.head())

if 'atac' in mdata.mod.keys():
    mdata['atac'].var_names_make_unique()
    if args.per_barcode_metrics_file is not None:
        L.info("Returning mudata with metadata from barcode metrics csv")
        per_barcode_metrics_file = pd.read_csv(args.per_barcode_metrics_file, index_col=0) #it's a csv by default and goes in atac obs
        mdata.mod['atac'].obs = mdata.mod['atac'].obs.merge(per_barcode_metrics_file, left_index=True, right_index=True)
    if args.peak_annotation_file is not None:
        L.info("Returning mudata with peak annotation in mdata['atac'].uns['atac']['peak_annotation']")
        add_peak_annotation(mdata, annotation=args.peak_annotation_file, sep="\t", return_annotation= False)
    if args.fragments_file is not None:
        L.info("Returning mudata with fragments file in mdata['atac'].uns['files']['fragments']")
        locate_fragments(mdata,fragments=args.fragments_file, return_fragments= False)


if 'rep' in mdata.mod.keys():
    col_update = mdata['rep'].obs.columns[mdata['rep'].obs.columns.str.contains("count")]
    mdata['rep'].obs[col_update] = mdata['rep'].obs[col_update].apply(pd.to_numeric)
    mdata.update()
    

#make var names unique
for mm in mdata.mod.keys():
    mdata[mm].var_names_make_unique()

mdata.obs['sample_id'] = str(args.sample_id)

# copy the  to each modality
for mm in mdata.mod.keys():
    print("saving sample_id to each modality")
    # mdata[mm].obs['sample_id'] = mdata.obs['sample_id']
    mdata[mm].obs['sample_id'] = mdata.obs.loc[mdata[mm].obs_names,:]['sample_id']

mdata.update()


L.info("saving data")
L.debug(mdata)
# this will write the file as anndata or muon depending on the class of mdata.
mdata.write(args.output_file)

L.info("done")




import argparse
import pandas as pd
import scanpy as sc
import numpy as np

from panpipes.funcs.processing import extract_parameter_from_fname, check_for_bool
from panpipes.funcs.io import read_anndata
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default="/well/dendrou/projects/cart/data/by_sample/pipe_qc/pipe_int_gut_pctmt50/pipe_clust_bbknn/gut_nneigh15_pcs50_dir/gut_bcbbknn_meteuclidean_neighbors.h5ad",
                    help='')
parser.add_argument('--output_anndata', default='adata_cellxgene.h5ad',
                    help='')
parser.add_argument('--use_muon',
                    default=False,
                    help='')
parser.add_argument('--modality',
                    default="rna",
                    help='')
parser.add_argument('--sample_prefix', default='adata_cellxgene.h5ad',
                    help='')               
parser.add_argument('--clusters', default="/well/dendrou/projects/cart/data/by_sample/pipe_qc/pipe_int_gut_pctmt50/pipe_clust_bbknn/gut_nneigh15_pcs50_dir/all_res_clusters_list.txt.gz",
                    help='')
parser.add_argument('--umaps', default="/well/dendrou/projects/cart/data/by_sample/pipe_qc/pipe_int_gut_pctmt50/pipe_clust_bbknn/gut_nneigh15_pcs50_dir/gut_md0.1_umap.txt.gz,/well/dendrou/projects/cart/data/by_sample/pipe_qc/pipe_int_gut_pctmt50/pipe_clust_bbknn/gut_nneigh15_pcs50_dir/gut_md0.5_umap.txt.gz",
                    help="")
parser.add_argument('--best_md', default=None)
parser.add_argument('--best_cluster_col', default=None)



args, opt = parser.parse_known_args()
L.info(args)
use_muon = check_for_bool(args.use_muon)

adata = read_anndata(args.infile, use_muon=use_muon, modality=args.modality)


# add in umap coordinates
umap_files = args.umaps.split(",")
umap_names = ["X_umap_md" + str(extract_parameter_from_fname(x, "md", "")) for x in umap_files]

for i in range(len(umap_files)):
    umap_coords = pd.read_csv(umap_files[i], sep='\t', index_col=0)
    if all(umap_coords.index == adata.obs_names):
        adata.obsm[umap_names[i]] = np.array(umap_coords)
    else:
        L.error("umap barcodes do not match anndata")

# add in clusters
clusters = pd.read_csv(args.clusters, sep='\t', index_col=0)
clusters = clusters.apply(lambda x: x.astype('category'))
if all(clusters.index == adata.obs_names):
    obs = adata.obs.merge(clusters, left_index=True, right_index=True)
    if all(obs.index == adata.obs_names):
        adata.obs = obs
else:
    L.error("cluster barcodes do not match anndata")

# move log counts to a layer
adata.layers['log_normcounts'] = adata.raw.X.copy()
adata.raw = None

# set favourtie params
if args.best_md is not None:
    try:
        adata.obsm['X_umap_int'] = adata.obsm['X_umap']
    except KeyError:
        L.info("integration umap not found, X_umap_int not created")
    adata.obsm['X_umap'] = adata.obsm['X_umap_md' + str(args.best_md)]

if args.best_cluster_col is not None:
    adata.obs['clusters'] = adata.obs[args.best_cluster_col]


adata.write(args.output_anndata)


L.info("Done")




import muon as mu
import scanpy as sc
import argparse
import pandas as pd
import os
from anndata import AnnData

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--infile", 
                    default="sc_preprocess.h5ad", help="file name, format: .h5ad")
parser.add_argument('--modality',
                    default=None,
                    help='')
parser.add_argument("--outfile", 
                    default="clusters.txt", help="file name, format: .txt")
parser.add_argument("--resolution",
                    default=0.5, help="no. neighbours parameters for sc.pp.neighbors()")
parser.add_argument("--algorithm", 
                    default="leiden", help="algortihm choice from louvain and leiden")
parser.add_argument("--neighbors_key", 
                    default="neighbors", help="algortihm choice from louvain and leiden")

args, opt = parser.parse_known_args()
L.info(args)

# read data
mdata = mu.read(args.infile)
if type(mdata) is AnnData:
    adata = mdata
elif args.modality is not None:
    adata = mdata[args.modality]
else:
    adata = mdata

uns_key=args.neighbors_key
# check sc.pp.neihgbours has been run
if uns_key not in adata.uns.keys():
    # sys.exit("Error: sc.pp.neighbours has not been run on this object")
    sc.pp.neighbors(adata)
    uns_key="neighbors"


# run command
if args.algorithm == "louvain":
    sc.tl.louvain(adata, resolution=float(args.resolution), key_added='clusters', neighbors_key=uns_key)
elif args.algorithm == "leiden":
    sc.tl.leiden(adata, resolution=float(args.resolution), key_added='clusters', neighbors_key=uns_key)
else:
    sys.exit("algorithm not found: please specify 'louvain' or 'leiden'")

#mdata.update()
## write out clusters as text file
clusters = pd.DataFrame(adata.obs['clusters'])
clusters.to_csv(args.outfile, sep='\t')

tmp = clusters['clusters'].value_counts().to_frame("cell_num").reset_index().rename(columns={"index":"cluster"})
tmp.to_csv(os.path.dirname(args.outfile) + "/cellnum_per_cluster.csv")

if "sample_id" in adata.obs.columns:
    ccounts = pd.DataFrame(adata.obs[['sample_id', 'clusters']])
    tmp = ccounts.value_counts().to_frame("cell_num").reset_index()
    tmp.to_csv(os.path.dirname(args.outfile) + "/cellnum_per_sample_id_per_cluster.csv")

L.info("Done")




# entry .py script adapted from COMBAT clusthub repo, originally by Adam Cribbs

'''
Panpipes - Single Cell analysis workflows 
============================================================

Please remember to acknowledge Charlotte Rich-Griffin and Fabiola Curion 
if you use this pipeline in your research

To use a specific workflow, type::
    panpipes <workflow> [workflow options] [workflow arguments]
For this message and a list of available keywords type::
    panpipes --help
To get help for a specify workflow, type::
    panpipes <workflow> --help
'''

import os
import sys
import re
import imp
import panpipes
from panpipes.R_scripts import __path__ as r_path
import subprocess

def main(argv=None):
    argv = sys.argv
    # paths to look for pipelines:
    path = os.path.abspath(os.path.join(os.path.dirname(panpipes.__file__), "panpipes"))
    relpath = os.path.abspath("./")
    paths = [path, relpath]
    if len(argv) == 1 or argv[1] == "--help" or argv[1] == "-h":
        print("The list of available workflows are:\n")
        pipelines_list = ['1. "ingest" : for the ingestion of data and computation of QC metrics' , 
                          '2. "preprocess" : for filtering and normalising of each modality',
                          '3. "integration" : integrate and batch correction using  single and multimodal methods', 
                          '4. "clustering" : cell clustering on single modalities', 
                          '5. "refmap" : transfer scvi-tools models from published data to your data', 
                          '6. "vis" : visualise metrics from other pipelines in context of experiment metadata',
                          '7. "qc_spatial" : for the ingestion of spatial transcriptomics (ST) data',
                          '8. "preprocess_spatial" : for filtering and normalizing ST data',
                          '9. "deconvolution_spatial" : for the cell type deconvolution of ST slides']
        print(*pipelines_list, sep="\n")
        return
    command = argv[1]
    if command != 'install_r_dependencies':
        command = re.sub("-", "_", command)
        pipeline = "pipeline_{}".format(command)
        del sys.argv[0]
        (file, pathname, description) = imp.find_module(pipeline, paths)
        module = imp.load_module(pipeline, file, pathname, description)
        module.main(sys.argv)
    else:
        subprocess.check_output(["Rscript", r_path[0] + "/install_R_libs.R"])



if __name__ == "__main__":
    sys.exit(main())



from collections import UserString
import sys
import warnings
import logging
import os 
import re
import numpy as np
import pandas as pd
import muon as mu

from pandas import concat
from pandas import merge as pdmerge
from itertools import compress, chain
from scanpy.get import obs_df
from scanpy.pp import subsample
from muon import MuData
from anndata import AnnData
# from muon.pp import intersect_obs
from random import sample
from functools import reduce


def is_float_try(string):
    """
    quick test to see if a string can be turned intoa float
    """
    try:
        float(string)
        return True
    except ValueError:
        return False


def splitall(path):
    """
    Function to completely split up a path into it's compoenents
    e.g. /path/to/cheese => ['path', 'to', 'cheese'] 
    inspired by https://stackoverflow.com/questions/3167154/how-to-split-a-dos-path-into-its-components-in-python
    """
    allparts = []
    # I don't know how this works but it does, magic?
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def extract_parameter_from_fname(fname, parameter, prefix):
    """
    Extract parameters from filename under the following circumstances:
    INPUT fname: res0.6_cluster.txt.gz or <parametervalue>_<analysis_type><file suffix>
    INPUT parameters: "res"
    OUTPUT: "0.6"
    Currently it doesn't work if the paramtervalue combo is next to the file suffix. (e.g. res0.6.txt.gz)
    """
    # first remove the sample prefix from the fname
    # so that any character strings that match the parameters are not picked up instead
    fname = re.sub("^" + prefix, "", fname)
    # split by underscores
    fname_split0 = [re.split('_', x) for x in splitall(fname)]  # split by path and _
    fname_split1 = list(chain.from_iterable(fname_split0))
    # extract the string that matches the requested parameter
    fname_grep = list(filter(lambda x: parameter in x, fname_split1))[0]
    # extract the specific value
    value = fname_grep.replace(parameter, "")
    # if the value is a number, convert to float or integer as appropriate
    if is_float_try(value):
        value = float(value)
        # convert integer values to integers e.g. 1.0 to 1 to match seurat columns
        if int(value) == float(value):
            value = int(value)
    return value

def intersect_obs_by_mod(mdata: MuData, mods:list =None):
    """
    Subset observations (samples or cells) in-place by intersect
    taking observations present only in modalities listed in mods, or all modalities if mods is None.
    Parameters
    ----------
    mdata: MuData
            MuData object
    """
    if mdata.isbacked:
        warnings.warn(
            "MuData object is backed. It might be required to re-read the object with `backed=False` to make the intersection work."
        )
    if mods is None:
        # we assume all modalities if mods is None
        mods = mdata.mod.keys()

    if len(mods) < 2:
        raise ValueError("cannot run intersect_obs on one modality")

    common_obs = reduce(np.intersect1d, [mdata[x].obs_names for x in mods])
    for mod in mdata.mod:
        mu.pp.filter_obs(mdata.mod[mod], common_obs)
    mdata.update_obs()
    return


def setdiff_obs_by_mod(mdata: MuData, x: str, y:str):
    """
    Subset observations (samples or cells) in-place by set difference
    taking observations present only in the mdata[x] and filtering mdata[y] to only contain the obs in x
    Use case example. You want to filter the rep modality to only include obsnames from rna, but you do not want
    to filter rna to only contain obs in the rep
    Parameters
    ----------
    mdata: MuData
            MuData object
    x:
    """
    if mdata.isbacked:
        warnings.warn(
            "MuData object is backed. It might be required to re-read the object with `backed=False` to make the intersection work."
        )
    
    # get the intersect between the two mods
    common_obs = reduce(np.intersect1d, [mdata[x].obs_names for x in [x,y]])
    # only filter the y mod
    mu.pp.filter_obs(mdata.mod[y], common_obs)
    mdata.update_obs()
    return


def downsample_adata(adata, nn, cat_col=None, seed=0):
    # if no category col, but adata in a list
    if cat_col is None:
        adatas = [adata]
    else:
        #just in case, remove any empty categories
        adata.obs[cat_col] = adata.obs[cat_col].cat.remove_unused_categories()
        # copy the obs for later
        # orig_obs = adata.obs.copy()
        adatas = [adata[adata.obs[cat_col].isin([clust])] for clust in adata.obs[cat_col].cat.categories]
    # do the subsample
    for dat in adatas:
        if dat.n_obs > nn:
            subsample(dat, n_obs=nn, random_state=seed)
    # the bit below is to be compatible with muon, otherwise we could just return adata_downsampled
    # get barcodes
    keep_bc = chain.from_iterable([list(x.obs_names) for x in adatas])
    adata = adata[list(keep_bc), ].copy()
    return(adata)


def downsample_mudata(mdata, nn, cat_col=None, mods: list = ['rna'], inplace=True, seed=0):
    # if inplace is False then we need to copy the object
    if inplace is False:
        mdata = mdata.copy()
    # do the downsample for 1 of the modalities.
    mdata.mod[mods[0]] = downsample_adata(mdata[mods[0]], cat_col=cat_col, nn=nn, seed=seed)
    if len(mods) > 1:
        intersect_obs_by_mod(mdata, mods=mods )
    return mdata




# def remove_double_TRB(filt_contig):
#     # remove clonotypes with more than one TRB
#     df = filt_contig[filt_contig['chain']=="TRB"].copy()
#     df['n_TRB'] = [x + 1 for x in list(df.groupby(['barcode']).cumcount())]
#     keep_bcs = list(df[df['n_TRB'] !=1].barcode.unique())
#     logging.info("removing %i annotations due to having more than 1 TRB chain" % len(keep_bcs))
#     filt_contig = filt_contig[~filt_contig["barcode"].isin(keep_bcs)]
#     return filt_contig, keep_bcs


# def define_clone_id(filt_contigs):
#     #filter out low confidence
#     df = filt_contigs[filt_contigs.is_cell & filt_contigs.high_confidence]
#     df = df[["barcode",'length', 'chain', 'v_gene', 'j_gene', 'c_gene', 'cdr3', 'cdr3_nt', "umis", "reads"]]
#     df['idx'] = [x + 1 for x in list(df.groupby(['barcode', 'chain']).cumcount())]
#     df['idx'] = df['idx'].astype("int64")
#     df['chain2'] = df['chain'] + df['idx'].astype("string")
#     df = df.pivot(index='barcode',columns='chain2')
#     df.columns = ['_'.join(col).strip() for col in df.columns.values]
#     df['clone_id'] = df.loc[:, df.columns.str.startswith('cdr3_TR')].fillna('').apply(lambda x: '-'.join(x),axis=1)
#     return df.loc[:,~df.columns.str.startswith("idx")]




def which_ind(bool_list):
    return list((compress(range(len(bool_list)), bool_list)))


def which_val(bool_list, val_list):
    return list((compress(val_list, bool_list)))


def check_for_bool(input_string):
    out = None
    if isinstance(input_string, str):
        if input_string == 'True':
            out = True
        elif input_string == 'False':
            out = False
        else:
            logging.info(input_string)
            raise TypeError("type error: please specify --%s as True or False" % input_string)
    elif isinstance(input_string, bool):
        return input_string
    else:
        # in this case the input is not a bool or a string
        raise TypeError("type error : please specify --%s as True or False (as a bool or a string)" %  input_string)
    return out

def test_file_or_value(test):
    try:
        float(test)
        out = "value"
    except:
        if os.path.isfile(test):
            out = "file"
        else:
            raise ValueError("no file or value found")
    return out

# this and the next function are very similar in that they essentialy are merge pandas functions
# either by index or by column.
def merge_with_adata_obs(adata, df, on_col="sample_id", inplace=False):
    """
    Merge a dataframe with adata.obs where the indexes do not match 
    between adata.obs and the df
    """
    if not isinstance(adata, AnnData):
        raise TypeError("adata is not an AnnData object")
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df is not an Pandas dataframe")
    if on_col not in adata.obs.columns:
        raise KeyError("%s is not in adata.obs.columns" % on_col)
    if on_col not in df.columns:
        raise KeyError("%s is not in df.columns" % on_col)
    # preserve the index
    adata.obs['barcode_id'] = adata.obs.index
    new_df = adata.obs.merge(df, how="left", on=on_col)
    # reenter index
    new_df.index = new_df['barcode_id']
    new_df = new_df.drop(columns=['barcode_id'])
    new_df.index.name = None
    if inplace:
        adata.obs = new_df
    else:
        return new_df


def add_var_mtd(prot, df, left_on="prot_id", right_on="prot_id"):
    if not isinstance(prot, AnnData):
        raise TypeError("prot is not an AnnData object")
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df is not an Pandas dataframe")
    # if the column we need is currently the index, make it a column
    if df.index.name == right_on:
        df[right_on] = df.index
        df.index.name=None
    # if the indexes of the adata and the df are equivalent go ahead with merge
    if len(set(prot.var[left_on]).intersection(set(df[right_on]))) == len(df[right_on]):
        # merge using indexes keeping only the vars from the orig prot
        prot_var = prot.var.merge(df, 
        left_on=left_on, right_on=right_on, 
        how="left")
        # check that the order is still ok (although the new df has no index, 
        # so we need to drop the index on the prev one.)
        if prot_var[left_on].equals(prot.var[left_on].reset_index(drop=True)):
            orig_index= prot.var.index
            prot.var = prot_var
            prot.var.index = orig_index
            logging.debug(prot.var.head())
        else:
            warnings.warn(UserWarning("prot.var is at risk of getting disordered, not merging the df"))
    



def update_var_index(adata, new_index_col):
    adata.var = adata.var.reset_index().set_index(new_index_col)


def concat_adatas(adatas, batch_key, join_type="inner"):
    if isinstance(adatas, list) and len(adatas)!=1:
        # L.info("concatenating ...")
        # pull out the sample names in the correct order for the adaatas list
        batches = [x.obs[batch_key][0] for x in adatas]
        adata = adatas[0].concatenate(adatas[1:], 
                                        batch_key=batch_key, 
                                        batch_categories=batches, 
                                        join=join_type)
    elif isinstance(adatas, list) and len(adatas)==1:
        adata = adatas[0]
    else:
        raise TypeError("adatas is not a list")
    return(adata)


def concat_mdatas(mdata_list, batch_key, join_type="inner"):
    logging.debug(batch_key)
    if len(mdata_list)!=1:
        slots_filled = set(chain(*[list(m.mod.keys()) for m in mdata_list]))
        logging.debug(slots_filled)
        # need to concatenate each slot separately then recreate mdata
        concat_adatas={}
        for sf in slots_filled:
            logging.debug(sf)
            # extract if modality exists (it's possuible that not all samples have all modalities (thinking specifically of rep))
            adata_list = [x[sf]for x in mdata_list if sf in x.mod.keys()]
            [x.var_names_make_unique() for x in adata_list]
            batches = [x.obs[batch_key][0] for x in adata_list]
            concat_adatas[sf] = adata_list[0].concatenate(adata_list[1:],
                                    batch_key=batch_key,
                                    batch_categories=batches, 
                                    join=join_type)
            
        mdata = MuData(concat_adatas)
        # for sf in slots_filled:
        #     mdata[sf].obs[batch_key] ==concat_adatas[sf].obs[batch_key]
        
        # make sure sample id is in the top obs (updated to not put extra rep columns in top obs)
        # changed to deal with circumstances where a category is missing from one modality
        all_sample_id_df = mdata.obs.iloc[:,mdata.obs.columns.str.endswith(":sample_id")]
        # first it needs to not be a category for the merge.
        all_sample_id_df = all_sample_id_df.apply(lambda x: x.astype('object'))
        mdata.obs['sample_id']  = all_sample_id_df.fillna('').astype(str).apply(lambda x: ' '.join(set(' '.join(x).split())), axis=1)
        # then turn it back into a category
        mdata.obs['sample_id'] = mdata.obs['sample_id'].astype('category')
        return mdata
    else:
        return mdata_list[0]

def concat_adata_list(adata_list, use_muon, **kwargs):
    if use_muon:
        out = concat_mdatas(adata_list, **kwargs)
    else:
        out = concat_adatas(adata_list, **kwargs)
    return out

def dedup(value):
    words = set(value.split(' '))
    return ' '.join(words)


def intersection(lst1, lst2): 
    lst3 = [value for value in lst1 if value in lst2] 
    return lst3 


def remove_unused_categories(df: pd.DataFrame):
    if not isinstance(df, pd.DataFrame):
        raise TypeError('df is not a pandas dataframe')
    for c in df.columns:
        if pd.api.types.is_categorical_dtype(df[c]):
            df[c] = df[c].cat.remove_unused_categories()
    

def mu_get_obs(mdata, features=[],modalities=[], layers=None):
    """
    returns pandas dataframe of features, having searched all layers for said features
    """
    if modalities is None:
        # assume we want all modalities
        mods = mdata.mod.keys()
    else:
        mods = modalities
    if layers is None:
        # we assume we want the X layer
            mod_dict = dict(zip(mods, [None]*len(mods)))
    else:
        if len(mods) == len(layers):
            mod_dict = dict(zip(mods, layers))
        else:
            logging.info("make sure mods is the same length as layers")

    out = {}
    for mod, layer in mod_dict.items():
        out[mod] = obs_df(mdata[mod], keys=features, layer=layer)
    df = concat(out, axis=1)
    df.columns = ['-'.join(col).strip() for col in df.columns.values]
    return df



import argparse
import os
import numpy as np
import scanpy as sc
import muon as mu
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
import panpipes.funcs as pnp

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("--filtered_mudata",
                    default=None,
                    help="input mudata after rna filtering(if rna in modality dictionary)")  
parser.add_argument("--bg_mudata",
                    default=None,
                    help="the mdata containing the raw measurements")
parser.add_argument("--channel_col",
                    default=None,
                    help="")
parser.add_argument("--normalisation_methods", default="clr,dsb",
                   help="comma separated list of normalisation methods")
parser.add_argument("--clr_margin",
                    default=0,
                    help="run clr with margin calculated on cells (columns) or on features (rows)")    
parser.add_argument("--quantile_clipping",
                    default=False,
                    help="")  
parser.add_argument("--store_as_x",
                    default=None,
                    help="which normalization to store in the X slot. Default to dsb if specified in normalization method")   
parser.add_argument("--figpath",
                    default="./prot_figures",
                    help="where to save the files")  
parser.add_argument("--save_mtx",
                    default=False,
                    help="if per channel normalization is applied, where to save the mtx file")  
parser.add_argument("--save_mudata_path",
                    default=None,
                    help="")   
parser.add_argument("--run_pca",
                    default=False,
                    help="whether to run pca on protein modality")   
parser.add_argument("--n_pcs",
                    default=None,
                    help="number of components to calculate")
parser.add_argument("--pca_solver",
                    default="arpack",
                    help="which PCA solver to use")
parser.add_argument("--color_by",
                    default="sample_id",
                    help="which columns to fetch from the protein .obs slot")


args, opt = parser.parse_known_args()
# args = argparse.Namespace(filtered_mudata='test.h5mu', 
# bg_mudata='/well/cartography/users/zsj686/non_cart_projects/005-multimodal_scpipelines/ingest/test_raw.h5mu', 
# channel_col=None, normalisation_methods='clr,dsb', clr_margin='0', quantile_clipping='True', figpath='./figures/prot', save_mtx=False, save_mudata_path='test.h5mu')
save_mtx=pnp.pp.check_for_bool(args.save_mtx)

norm_methods = args.normalisation_methods.split(',')


L.info(args)

figdir = args.figpath
if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))



# load filtered data - if use_muon is True this will return a mudata object, else returns an mudata object

L.info("reading filtered mudata object")
try:
    all_mdata = mu.read(args.filtered_mudata)
except FileNotFoundError:
    sys.exit("filtered_mudata file not found")

# load bg data
if 'dsb' in norm_methods:
    if args.bg_mudata is not None:
        L.info("reading bg file object")
        try:
            all_mdata_bg = mu.read(args.bg_mudata)
            # subset to just the channel 
        except FileNotFoundError:
            sys.exit("bg_mudata object not found")
    else:
        sys.exit("must specify a bg mudata to run dsb, containing both rna and prot")
    
    # checking that the samne proteins are in foreground and background (since foreground might have been filtered)
    if len(all_mdata['prot'].var_names) != len(all_mdata_bg['prot'].var_names):
        mu.pp.filter_var(all_mdata_bg['prot'], all_mdata['prot'].var_names)
        all_mdata_bg.update()


# find isotypes columns 
if "isotype" in all_mdata["prot"].var.columns:
    isotypes = list(all_mdata["prot"].var_names[all_mdata["prot"].var.isotype])
    # exclude isotypes from downstream analysis by setting them as the only not highly variable genes.
    all_mdata['prot'].var['highly_variable'] = ~all_mdata['prot'].var['isotype'] 
else:
    isotypes = None
    
L.info("isotypes found:")
L.info(isotypes)


# save raw counts
if pnp.scmethods.X_is_raw(all_mdata["prot"]):
    all_mdata["prot"].layers["raw_counts"] = all_mdata["prot"].X.copy()
elif "raw_counts" in all_mdata["prot"].layers :
    L.info("raw_counts layer already exists")
    all_mdata.X = all_mdata["prot"].layers['raw_counts'].copy()
else:
    L.error("X is not raw data and raw_counts layer not found")
    sys.exit("X is not raw data and raw_counts layer not found")



if args.channel_col is not None:
    ## running per channel (parallelise in future)
    channel_cols = list(pnp.processing.mu_get_obs( all_mdata, 
        features=[args.channel_col],
        modalities=["prot"]).iloc[:,0].unique())
    #this line works only if the obs of the entire metadata is the same as the prot obs. which may not always be the case
    if not args.channel_col in all_mdata.obs.columns:
        print("adding channel col to mdata obs since it's only in the prot obs")
        all_mdata.obs.loc[all_mdata["prot"].obs.index,args.channel_col] = all_mdata["prot"].obs[args.channel_col]
    mdata_per_sample = {si: all_mdata[all_mdata.obs[args.channel_col] == si] for si in channel_cols}
    # this data is only necessary if running dsb
    if 'dsb' in norm_methods:
        mdata_bg_per_sample = {si: all_mdata_bg[all_mdata_bg.obs[args.channel_col] == si] for si in channel_cols}
    for si in channel_cols:
        # si = list(mdata_per_sample.keys())[0]
        L.info(si)
        mdata=mdata_per_sample[si]
        plot_features = list(mdata["prot"].var_names)
        plot_features.sort()
        plot_features = [x for x in plot_features if x not in isotypes]
        # first run clr
        if 'clr' in norm_methods:
            # make sure to start from raw counts
            mdata["prot"].X = mdata["prot"].layers["raw_counts"].copy()
            pnp.scmethods.run_prot_normalise(mdata=mdata, mdata_bg=None,
                    method="clr",
                    clr_margin=int(args.clr_margin))
            L.info("saving ridgeplot")
            pnp.plotting.ridgeplot(mdata["prot"], features=plot_features, layer="clr",  splitplot=6)
            plt.savefig(os.path.join(args.figpath, str(si) + "_clr_ridgeplot.png"))
            if isotypes is not None:
                pnp.plotting.ridgeplot(mdata["prot"], features=isotypes, layer="clr",  splitplot=1)
                plt.savefig(os.path.join(args.figpath, str(si) + "_clr_ridgeplot_isotypes.png"))
            # save out the data in what format?
            if save_mtx:
                pnp.io.write_10x_counts(mdata["prot"], os.path.join("prot_clr" , str(si) ), layer="raw_counts")
        # then run dsb
        if 'dsb' in norm_methods:
            mdata_bg = mdata_bg_per_sample[si]
            mdata = mdata_per_sample[si]
            # make sure to start from raw counts
            mdata["prot"].X = mdata["prot"].layers["raw_counts"].copy()
            # plot hiustogram of bg
            mdata_bg["rna"].obs["log10umi"] = np.array(np.log10(mdata_bg["rna"].X.sum(axis=1) + 1)).reshape(-1)
            # mu.pl.histogram(mdata_bg["rna"], ["log10umi"], bins=50)
            # plt.savefig(os.path.join(args.figpath, si + "_log10umi.png"))
            # run dsb
            # this stores a layer named after the method as well as overwriting X
            pnp.scmethods.run_prot_normalise(mdata=mdata, 
                                            mdata_bg=mdata_bg,
                                            method="dsb", 
                                            isotypes=isotypes)
            pnp.plotting.ridgeplot(mdata["prot"], features=plot_features, layer="dsb",  splitplot=6)
            plt.savefig(os.path.join(args.figpath, str(si) + "_dsb_ridgeplot.png"))
            if isotypes is not None:
                pnp.plotting.ridgeplot(mdata["prot"], features=isotypes, layer="dsb",  splitplot=1)
                plt.savefig(os.path.join(args.figpath, str(si) + "_dsb_ridgeplot_isotypes.png"))
            if save_mtx:
                pnp.io.write_10x_counts(mdata["prot"], os.path.join("prot_dsb" , str(si)), layer="raw_counts")
else:
    # run on all the data (not on a channel basis)
    # first run clr
     # do the ridgeplots for clr
    plot_features = list(all_mdata["prot"].var_names)
    plot_features.sort()
    plot_features = [x for x in plot_features if x not in isotypes]
    if 'clr' in norm_methods:
        # make sure to start from raw counts
        all_mdata["prot"].X = all_mdata["prot"].layers["raw_counts"].copy()
        # run normalise
        # this stores a layer named after the method as well as overwriting X
        pnp.scmethods.run_prot_normalise(mdata=all_mdata, 
                mdata_bg=None,
                method="clr",
                clr_margin=int(args.clr_margin))
        pnp.plotting.ridgeplot(all_mdata["prot"], features=plot_features, layer="clr",  splitplot=6)
        plt.savefig(os.path.join(args.figpath, "clr_ridgeplot.png"))
        if isotypes is not None:
            pnp.plotting.ridgeplot(all_mdata["prot"], features=isotypes, layer="clr",  splitplot=1)
            plt.savefig(os.path.join(args.figpath, "clr_ridgeplot_isotypes.png"))
        # save out the data in what format?
        if save_mtx:
            pnp.io.write_10x_counts(all_mdata["prot"], os.path.join("prot_clr"), layer="clr")
    if 'dsb' in norm_methods:
        # make sure to start from raw counts
        all_mdata["prot"].X = all_mdata["prot"].layers["raw_counts"].copy()
        # then run dsb
        # comput log 10 umi
        all_mdata_bg["rna"].obs["log10umi"] = np.array(np.log10(all_mdata_bg["rna"].X.sum(axis=1) + 1)).reshape(-1)
        mu.pl.histogram(all_mdata_bg["rna"], ["log10umi"], bins=50)
        plt.savefig(os.path.join(args.figpath, "all_log10umi.png"))
        # this stores a layer named after the method as well as overwriting X
        pnp.scmethods.run_prot_normalise(mdata=all_mdata, 
                mdata_bg=all_mdata_bg, 
                method="dsb",
                isotypes=isotypes) 
        # apply quantile clipping # discussed in FAQs https://cran.r-project.org/web/packages/dsb/vignettes/dsb_normalizing_CITEseq_data.html
        if args.quantile_clipping:
            all_mdata['prot'].layers['dsb'] = pnp.scmethods.quantile_clipping(all_mdata['prot'], layer="dsb", inplace=False)
            all_mdata['prot'].X = all_mdata['prot'].layers['dsb'].copy()
        # plot ridgeplot
        pnp.plotting.ridgeplot(all_mdata["prot"], features=plot_features, layer="dsb",  splitplot=6)
        plt.savefig(os.path.join(args.figpath, "dsb_ridgeplot.png"))
        if isotypes is not None:
            pnp.plotting.ridgeplot(all_mdata["prot"], features=isotypes, layer="dsb",  splitplot=1)
            plt.savefig(os.path.join(args.figpath, "dsb_ridgeplot_isotypes.png"))
        # save out the data in what format?
        if save_mtx:
            pnp.io.write_10x_counts(all_mdata["prot"], os.path.join("prot_dsb"), layer="dsb")
    
    if args.store_as_x is not None:
        all_mdata["prot"].X = all_mdata["prot"].layers[args.store_as_x]
        
    # run pca on X 
    # this basically makes no sense if you have a small panel of antibodies.
    if args.run_pca:
        L.info("running pca")

        if all_mdata['prot'].var.shape[0] < int(args.n_pcs):
            L.info("You have less features than number of PCs you intend to calculate")
            n_pcs = all_mdata['prot'].var.shape[0] - 1
            L.info("Setting n PCS to %i" % int(n_pcs))
        else:
            n_pcs = int(args.n_pcs)
        sc.tl.pca(all_mdata['prot'], n_comps=n_pcs, 
                        svd_solver=args.pca_solver, 
                        random_state=0) 

        # do some plots!
        sc.pl.pca_variance_ratio(all_mdata['prot'], log=True, n_pcs=n_pcs, save=".png")
        
        L.info(args.color_by)
        
        col_variables = args.color_by.split(",")
        L.info("col_variables")
        L.info(col_variables)
        
        col_variables = [a.strip() for a in col_variables]

        L.info("col_variables now")
        L.info(col_variables)
        
        col_use = [var for var in col_variables if var in all_mdata['prot'].obs.columns]
        L.info("col_use")
        L.info(col_use)
        sc.pl.pca(all_mdata['prot'], color=col_use, save = "_vars.png")
        sc.pl.pca_loadings(all_mdata['prot'], components="1,2,3,4,5,6", save = ".png")
        sc.pl.pca_overview(all_mdata['prot'], save = ".png")

    if args.save_mudata_path is not None:
        all_mdata.update()
        all_mdata.write(args.save_mudata_path)


L.info("done")



import pandas as pd
import numpy as np
# from scanpy.pp import normalize_total
from scipy.sparse import issparse
from scanpy.get import obs_df as get_obs_df
from scanpy.pp import normalize_total
import scanpy as sc
import warnings
import logging
from typing import Optional, Literal
import sys
import muon as mu
from muon import MuData
from anndata import AnnData
import os
from matplotlib.pyplot import savefig
from statsmodels.distributions.empirical_distribution import ECDF

from .plotting import ridgeplot
from .io import write_10x_counts
from .processing import check_for_bool, which_ind
import matplotlib
import matplotlib.pyplot as plt


def lsi(adata, num_components):
    """
    Runs LSI on adata.X and uses HVFs if adata.var["highly_variable"] is present. Otherwise, runs LSI on all features.

    :param adata: anndata to run LSI on
    :param num_components: number of components to compute
    """

    if "highly_variable" in adata.var.columns:
        adata_hvfs = adata[:, adata.var["highly_variable"]]
        mu.atac.tl.lsi(adata_hvfs, n_comps=num_components)
        # save output to original anndata
        adata.obsm["X_lsi"] = adata_hvfs.obsm['X_lsi']
        adata.uns["lsi"] = adata_hvfs.uns['lsi']
        adata.varm["LSI"] = np.zeros(shape=(adata.n_vars, adata_hvfs.varm["LSI"].shape[1]))
        adata.varm["LSI"][adata.var["highly_variable"]] = adata_hvfs.varm['LSI']
    else:
        mu.atac.tl.lsi(adata, n_comps=num_components)


def cell2loc_filter_genes(adata, fig_path, cell_count_cutoff=15, cell_percentage_cutoff2=0.05, nonz_mean_cutoff=1.12):
# Adjusted from: https://github.com/BayraktarLab/cell2location/blob/master/cell2location/utils/filtering.py
    adata.var["n_cells"] = np.array((adata.X > 0).sum(0)).flatten()
    adata.var["nonz_mean"] = np.array(adata.X.sum(0)).flatten() / adata.var["n_cells"]

    cell_count_cutoff = np.log10(cell_count_cutoff)
    cell_count_cutoff2 = np.log10(adata.shape[0] * cell_percentage_cutoff2)
    nonz_mean_cutoff = np.log10(nonz_mean_cutoff)

    gene_selection = (np.array(np.log10(adata.var["n_cells"]) > cell_count_cutoff2)) | (
        np.array(np.log10(adata.var["n_cells"]) > cell_count_cutoff)
        & np.array(np.log10(adata.var["nonz_mean"]) > nonz_mean_cutoff)
    )
    gene_selection = adata.var_names[gene_selection]
    adata_shape = adata[:, gene_selection].shape

    fig, ax = plt.subplots()
    ax.hist2d(
        np.log10(adata.var["nonz_mean"]),
        np.log10(adata.var["n_cells"]),
        bins=100,
        norm=matplotlib.colors.LogNorm(),
        range=[[0, 0.5], [1, 4.5]],
    )
    ax.axvspan(0, nonz_mean_cutoff, ymin=0.0, ymax=(cell_count_cutoff2 - 1) / 3.5, color="darkorange", alpha=0.3)
    ax.axvspan(
        nonz_mean_cutoff,
        np.max(np.log10(adata.var["nonz_mean"])),
        ymin=0.0,
        ymax=(cell_count_cutoff - 1) / 3.5,
        color="darkorange",
        alpha=0.3,
    )
    plt.vlines(nonz_mean_cutoff, cell_count_cutoff, cell_count_cutoff2, color="darkorange")
    plt.hlines(cell_count_cutoff, nonz_mean_cutoff, 1, color="darkorange")
    plt.hlines(cell_count_cutoff2, 0, nonz_mean_cutoff, color="darkorange")
    plt.xlabel("Mean non-zero expression level of gene (log)")
    plt.ylabel("Number of cells expressing gene (log)")
    plt.title(f"Gene filter: {adata_shape[0]} cells x {adata_shape[1]} genes")
    plt.savefig(fig_path)

    return gene_selection




def findTopFeatures_pseudo_signac(adata, min_cutoff):
    # Adapted from:
    # https://stuartlab.org/signac/reference/findtopfeatures
    """
    :param adata:
    :param min_cutoff:
        "q[x]": "q" followed by the minimum percentile, e.g. q5 will set the top 95% most common features as higly variable
        "c[x]": "c" followed by a minimum cell count, e.g. c100 will set features present in > 100 cells as highly variable
        "tc[x]": "tc" followed by a minimum total count, e.g. tc100 will set features with total counts > 100 as highly variable
        "NULL": All features are assigned as highly variable
        "NA": Highly variable features won't be changed
    :return: Percentile rank of each feature stored in "adata.var["percentile"], highly variable features stored in adata.var["highly_variable"].
    "total_counts" and "n_cells_by_counts" also saved, if not present.
    Filtering is done according to the "min_cutoff" parameter. If min_cutoff == NA, adata.var["highly_variable"] won't be changed
    """
    if "total_counts" not in adata.var:
        adata.var["total_counts"] = np.ravel(adata.layers["raw_counts"].sum(axis=0))
    if "n_cells_by_counts" not in adata.var:
        if issparse(adata.layers["raw_counts"]):
            adata.var["n_cells_by_counts"] = adata.layers["raw_counts"].getnnz(axis=0)
        else:
            adata.var["n_cells_by_counts"] = np.count_nonzero(adata.layers["raw_counts"], axis=0)


    if "percentile" not in adata.var:
        # fit ecdf
        ecdf = ECDF(adata.var["total_counts"])
        extracted = adata.var[["total_counts"]]
        extracted = extracted.sort_values("total_counts", ascending=False)
        # add decreasingly ordered percentiles to dataframe:
        extracted["percentile"] = -np.sort(-ecdf.y)[:-1]  # last percentile always corresponds to the percentile of -Inf
        # merge with adata.var
        adata.var = pd.merge(adata.var, extracted.drop("total_counts", axis=1), left_index=True, right_index=True)

    if min_cutoff.startswith("q"):  # quantile filtering
        min_cutoff = int(min_cutoff[1:]) / 100
        adata.var['highly_variable'] = [True if percentile > min_cutoff else False for percentile in adata.var["percentile"]]
        return
    if min_cutoff.startswith("c"):  # filtering by minimum number of cells
        min_cutoff = int(min_cutoff[1:])
        adata.var['highly_variable'] = [True if ncells > min_cutoff else False for ncells in adata.var["n_cells_by_counts"]]
        return
    if min_cutoff.startswith("tc"):  # filtering by total counts
        min_cutoff = int(min_cutoff[2:])
        adata.var['highly_variable'] = [True if totalcounts > min_cutoff else False for totalcounts in adata.var["total_counts"]]
        return
    if min_cutoff == "NA":  # don't change variable features
        return
    if min_cutoff == "NULL":  # assign all features as highly variable
        adata.var["highly_variable"] = True
        return

def extract_lsi(adata):
    if "X_lsi" not in adata.obsm.keys():
        sys.error("Can't extract lsi from this object")
    else:
        tab = pd.concat([
        pd.DataFrame(
        adata.obsm["X_lsi"],
        index=adata.obs_names,
        columns=[f"LSI_{i+1}" for i in range(adata.obsm["X_lsi"].shape[1])]),
        adata.obs,
        ],axis=1)
        return tab

def calc_tech_corr(lsi_df,tech_covariates=['n_genes_by_counts','total_counts'], ncomps = 20):
    """
    lsi_df = extract_lsi(adata)
    tech_covariates = what columns of the obs to use
    ncomps = which LSI components to show
    """
    lsi_columns = [f"LSI_{i}" for i in range(1, ncomps+1)]  # Top 20 LSI columns
    correlation_data = []

    for col in lsi_columns:
        for tech in tech_covariates:
            corr_value = lsi_df[col].corr(lsi_df[tech])
            correlation_data.append({'LSI_Component': col, 'Tech_Covariate': tech, 'Correlation': corr_value})

    # Creating a DataFrame
    correlation_df = pd.DataFrame(correlation_data)
    return correlation_df
    
    

def exp_mean_sparse(x):
    """
    This returns the log of the mean of the not-logged data
    this version of the function works directly on the sparse matrix but is slow.
    """
    # convert out of compressed sparse matrix
    return np.log(x.expm1().mean(1)+1).T


def exp_mean_dense(x):
    """
    This returns the log of the mean of the not-logged data
    this version of the function requires a dense matrix, so might be memory hungry?
    But it is super fast
    """
    # convert out of compressed sparse matrix
    return np.log((np.sum(np.exp(x)-1)/x.shape[1]) + 1)

def find_all_markers_pseudo_seurat(
        adata, 
        groups,
        groupby,
        layer=None,
        method=None,
        n_genes=float("inf"), 
        corr_method="bonferroni",
        arg_minpct=0.1,
        arg_mindiffpct=-float("inf"), 
        arg_logfcdiff=0.25):
    # add replace X with layer
    if layer is not None:
        adata.X = adata.layers[layer]
    # need to check is the assay layer is dense or not
    assay_is_sparse = issparse(adata.X)
    use_dense = assay_is_sparse==False 
    if groups == 'all':
        groups = adata.obs[groupby].unique().tolist()
    markers_dict = {}
    filter_dict = {}
    for cv in groups:
        # \ set up idenst as cv ==1 and everything else = 0
        adata.obs['idents'] = ['1' if x == cv else '0' for x in adata.obs[groupby]]
        filter_dict[cv] = pseudo_seurat(adata, use_dense=use_dense,arg_minpct=arg_minpct,
                  arg_mindiffpct=arg_mindiffpct, 
                  arg_logfcdiff=arg_logfcdiff )
        logging.info("number of genes remaining after filtering:  %i\n" % filter_dict[cv]['background'].sum())
        adata_rg = adata[:, filter_dict[cv]['background'].tolist()].copy()
        sc.tl.rank_genes_groups(adata_rg, layer=layer,
                                groupby="idents", groups=["1"],  
                                reference="0",
                                method=method, 
                                n_genes=float("inf"), 
                                corr_method="bonferroni")
        markers_dict[cv] = sc.get.rank_genes_groups_df(adata_rg, group="1")
        # remove adata from mem
        adata_rg = None
    markers = pd.concat(markers_dict.values(), keys=markers_dict.keys())
    filter_stats = pd.concat(filter_dict.values(), keys=filter_dict.keys())
    return markers, filter_stats

def pseudo_seurat(adata, 
                  arg_minpct=0.1,
                  arg_mindiffpct=-float("inf"), 
                  arg_logfcdiff=0.25, 
                  use_dense=False):
    """
    alternative method that"s more like seurat (pseudo seurat if you will)
    In that you filter genes before running rank genes
    ---
    1.  define idents
    2. define cluster cells and other cells
    3. check min cells
    4. compute percentages and difference from equiv data slot to s.obj@assays$RNA@data
    5. computer mean expression levels and differences
    6. save stats
    7. define background based on min_pct (pct of cells that have to express marker)
    8. filter stats for testing and save universe
    genes|cluster_mean|other_mean|diff_mean|cluster_pct|other_pct|max_pct|min_pct|diff_pct|background
    9. Filter adata.raw based on these genes?

    9. Find markers based on cluster cells and other cells,
    values from adata.raw ??
    gene: only expressed
    cells: we need to define manually based on above stats.
    10. save results.
    """
    # define cells
    cluster_cells_ind = which_ind(adata.obs["idents"] == "1")
    other_cells_ind = which_ind(adata.obs["idents"] == "0")

    # compute perecentage expressed
    # from normnalised but not scaled data
    # remember cells are rows and genes are columns

    # note: I don't know why norm_counts[cluster_cell_ind:, col_ind] deosn"t work, but it doesn't
    cluster_pct = (adata.X[cluster_cells_ind, :] > 0).sum(axis=0) / len(cluster_cells_ind)
    other_pct = (adata.X[other_cells_ind, :] > 0).sum(axis=0) / len(other_cells_ind)

    pcts = pd.DataFrame(np.vstack((cluster_pct,  other_pct)).transpose())
    max_pct = pcts.max(axis=1)
    min_pct = pcts.min(axis=1)
    diff_pct = max_pct - min_pct
    take_diff_pct = diff_pct > arg_mindiffpct
    # remove genes that are not expressed higher than 0.1 in one of the groups
    take_min_pct = max_pct > arg_minpct

    # KEEP IN CASE NP.ARRAY METHOD USES TOO MUCH MEMORY
    # import time
    # this has the potential to be very slow. Transposeing it speeds it up a bit.
    # I need to undertand sparse matrices better to make it work
    if use_dense:
        logging.info("using dense matrix")
        # extract the counts for cluster cells and calculate exp means on each row
        nct = adata.X.T[:, cluster_cells_ind]
        cluster_mean = np.apply_along_axis(exp_mean_dense, 1, nct.todense())

        # likewise for non-cluster cells
        nct = adata.X.T[:, other_cells_ind]
        other_mean = np.apply_along_axis(exp_mean_dense, 1, nct.todense())
        diff_mean = abs(cluster_mean - other_mean)
    else:
        logging.info("using sparse matrix")
        cluster_mean = exp_mean_sparse(adata.X.T[:, cluster_cells_ind])
        other_mean = exp_mean_sparse(adata.X.T[:, other_cells_ind])
        diff_mean = abs(cluster_mean - other_mean).A1

    # remove genes with less than threshold difference
    take_thresh = diff_mean > arg_logfcdiff
    # take = if a cell passes all the tests then it is to be kept.
    take = [a and b and c for a, b, c in zip(take_thresh, take_min_pct, take_diff_pct)]
    print("saving universe for fisher test")
    stats_df = pd.DataFrame(np.vstack((adata.var_names, cluster_mean, other_mean, diff_mean,
                                       cluster_pct, other_pct, max_pct, min_pct, diff_pct, take)).transpose(),
                            columns=["gene", "cluster_mean", "other_mean", "diff_mean",
                                     "cluster_pct", "other_pct",
                                     "max_pct", "min_pct", "diff_pct", "background"])
    return stats_df



def run_neighbors_method_choice(adata, method, n_neighbors, n_pcs, metric, use_rep, nthreads=1):
    # This works with both Anndata and MuData inputs 
    # useful if we are dealing with a MuData object but we want to use single rep, e.g.
    # calculating neighbors on a totalVI latent rep
    if n_pcs > adata.n_vars:
        logging.info("Reducing the number of components from %i to %i" %(n_pcs, adata.n_vars-1))
        n_pcs = adata.n_vars-1
    if method == "scanpy":
        logging.info("Computing neighbors using scanpy")
        from scanpy.pp import neighbors
        neighbors(adata,
                        n_pcs=n_pcs,
                        n_neighbors=n_neighbors,
                        metric=metric,
                        use_rep=use_rep)
    elif method == "hnsw":
        from scvelo.pp import neighbors
        logging.info("Computing neighbors using hnswlib (with scvelo a la pegasus!)")
        # we use the neighbors function from scvelo (thanks!)
        # with parameters from pegasus (for a more exact result).
        # code snippet from Steve Sansom, via COMBAT project
        neighbors(adata,
                n_neighbors=int(n_neighbors),
                n_pcs=n_pcs,
                use_rep=use_rep,
                knn=True,
                random_state=0,
                method='hnsw',
                metric=metric,
                metric_kwds={"M": 20,
                            "ef": 200,
                            "ef_construction": 200},
                num_threads=int(nthreads))


def merge_consensus_clust(adata, consensus_clust, ref_col="rough_ref"):
    if ref_col in adata.obs.columns:
        out = adata
    else:
        xx = consensus_clust[consensus_clust.index.isin(adata.obs_names)]
        adata.obs = adata.obs.merge(xx,how="left", left_index=True, right_index=True)
        adata.obs[ref_col] = adata.obs[ref_col].astype('category') 
        out = adata
    return(out)


# assessing background

def _calc_top_n_genes(adata, n_top=50):
    norm_dict = normalize_total(adata, target_sum=100, inplace=False)
    
    if issparse(norm_dict['X']):
            mean_percent = norm_dict['X'].mean(axis=0).A1
            top_idx = np.argsort(mean_percent)[::-1][:n_top]
            counts_top_genes = norm_dict['X'][:, top_idx].A
    else:
        mean_percent = norm_dict['X'].mean(axis=0)
        top_idx = np.argsort(mean_percent)[::-1][:n_top]
        counts_top_genes = norm_dict['X'][:, top_idx]
    columns = (
        adata.var_names[top_idx]
    )
    counts_top_genes = pd.DataFrame(
        counts_top_genes, index=adata.obs_names, columns=columns
    )
    
    return list(counts_top_genes.columns)

def get_top_expressed_features(adata, n_top=50, group_by=None):
    """
    Get highest expressed dfeatures from an andata object.
    If group by is specfied the union of the top features 
    from each grouping var will be returned
    
    Parameters
    ----------
    adata : AnnData object
    n_top : int
              No. of features to return (per grouping)
    group_by : str
              Designates a column from adata.obs on which to group
              vars.
    """
    if group_by is not None:
        groups = adata.obs[group_by].unique()
        top_genes_list = [_calc_top_n_genes(adata[adata.obs[group_by] == x, :], n_top=n_top) for x in groups]
        from itertools import chain
        top_genes = list(set(chain.from_iterable(top_genes_list)))
    else:
        top_genes = _calc_top_n_genes(adata, n_top=n_top)
    return(top_genes)

    
def get_mean_background_fraction(adata, top_background, group_by=None):
    norm_dict = normalize_total(adata, target_sum=100, inplace=False)
    gene_idx = [idx for idx, gene in  enumerate(adata.var_names ) if gene in top_background]
    if issparse(norm_dict['X']):
        mean_percent = norm_dict['X'].mean(axis=0).A1
        counts_top_genes = norm_dict['X'][:, gene_idx].A
    else:
        mean_percent = norm_dict['X'].mean(axis=0)
        counts_top_genes = norm_dict['X'][:, gene_idx]

    columns = (
        adata.var_names[gene_idx]
    )
    counts_top_genes = pd.DataFrame(
        counts_top_genes, index=adata.obs_names, columns=columns
    )
    if group_by is None:
        means_df = counts_top_genes.mean()
    else:
        obs_df = get_obs_df(adata, [group_by])
        counts_top_genes = pd.merge(counts_top_genes, obs_df, left_index=True, right_index=True)
        means_df = counts_top_genes.groupby(group_by).mean()
    return means_df


def identify_isotype_outliers(prot, isotypes, quantile_val=0.9, n_isotypes_pass=2, groupby=None,layer=None, inplace=True):
    # make sure we don't mess
    if inplace is False:
        prot=prot.copy()
    if groupby is None:
            isotypes_counts = get_obs_df(prot, isotypes, layer=layer)
            # get quantile thresholds
            thres = isotypes_counts.quantile(quantile_val)
            # for each isotype test whether the threshold is exceeded
            for iso in isotypes:
                isotypes_counts['outliers_' + iso]= isotypes_counts[iso] > thres[iso] 
    else:
        # we are grouping by an extra variable
        isotypes_counts = get_obs_df(prot, [groupby] + isotypes, layer=layer)
        thres = isotypes_counts.groupby(groupby).quantile(0.9)
        for iso in isotypes:
            tmp = pd.merge(isotypes_counts[[groupby, iso]], thres[[iso]], left_on=groupby, right_index=True)
            isotypes_counts['outliers_' + iso] = tmp.apply(lambda x: x[1] > x[2] , axis=1)
    # now work out which ones to exclude
    isotypes_counts['outliers'] = isotypes_counts.iloc[:,isotypes_counts.columns.str.startswith("outliers")].sum(axis=1)
    isotypes_counts['exclude'] = isotypes_counts['outliers'] > n_isotypes_pass
    # add into obs
    prot.obs['isotype_exclude_outliers'] = isotypes_counts['exclude']
    prot.obs['n_isotype_in_90_percentile'] = isotypes_counts['outliers'].astype('category')
    if inplace is False:
        return prot

def X_is_raw(adata):
    # we assume if all the rows sums are integers then the matrix is populated by integers
    return np.array_equal(adata.X.sum(axis=0).astype(int), adata.X.sum(axis=0))

# run clr 
    

def run_prot_normalise(mdata, 
                      mdata_bg, 
                      method: Literal['dsb', 'clr'] = 'clr',
                      clr_margin: Literal[0, 1] = '0',
                      isotypes=None):
    # rplace bg counts back in X
    # check if integers in mdata['prot'].X
    if X_is_raw(mdata['prot']) is False:
        raise ValueError("mdata['prot'].X does not contain raw counts, cannot normalise")
    if method == "clr":
        mu.prot.pp.clr(mdata["prot"], inplace=True, axis=int(clr_margin))
        mdata["prot"].layers["clr"] = mdata["prot"].X.copy()
    elif method == "dsb":
        # check mdata_bg actually contains raw counts
        mu.prot.pp.dsb(mdata,
                    mdata_bg, 
                    empty_counts_range=(1.0, 2.7), 
                    isotype_controls=isotypes, 
                    random_state=4)

        mdata["prot"].layers["dsb"] = mdata["prot"].X.copy()
    else:
        raise ValueError("normalisation method %s not recognised, please choose dsb or clr" % method)
    

from scipy.sparse import issparse
def quantile_clipping(mdata, modality=None, layer=None,  qmax=0.995, qmin=0.005, inplace=False):
    if isinstance(mdata, MuData):
        # need to pull out the modality and layer
        if modality is None:
            raise ValueError("if passing a MuData object, please specify a modality name e.g. prot")
        if layer is None:
            arr = mdata[modality].X
        else:
            arr = mdata[modality].layers[layer]
        if issparse(arr):
            arr = arr.todense()
        df = pd.DataFrame(data=arr, index=mdata[modality].obs_names.tolist(), columns=mdata[modality].var_names.tolist())
    elif isinstance(mdata, AnnData):
        if layer is None:
            arr = mdata.X
        else:
            arr = mdata.layers[layer] 
        if issparse(arr):
            arr = arr.todense()
        df = pd.DataFrame(data=arr, index=mdata.obs_names.tolist(), columns=mdata.var_names.tolist())
    # do quantile clipping
    qlow = np.quantile(arr, qmin, axis=0)
    qhigh = np.quantile(arr, qmax, axis=0)  
    df_clipped = df.clip(lower = qlow, upper = qhigh, axis=1).to_numpy()
    # return results
    if inplace:
        if isinstance(mdata, MuData):
        # need to pull out the modality and layer
            if layer is None:
                mdata[modality].X = df_clipped
            else:
                mdata[modality].layers[layer] = df_clipped
        elif isinstance(mdata, AnnData):
            if layer is None:
                mdata.X = df_clipped
            else:
                mdata.layers[layer] = df_clipped
    else:
        return df_clipped
    